# @authors: Andrew P. Craig
# @email: apc53@cam.ac.uk
# @date: 2017/10/09
# @copyright: MIT License

# Creates directories containing files necessary to run on cluster or locally.

import numpy as np
import time
import datetime
import shutil, os
import glob
import hlbutils
import pandas as pd
import platform
import sys
import HLBPlantYears

inputDir = '../HLBAppsAndInput/'
allSimsDir = '../HLBSims/'

def makeJobsFromSimSummary(simSummaryID, message, useNewJobID=True):

    # Load simSummary.
    oldSimSummaryFile = 'ID_' + str(simSummaryID) + '/summary_ID_' + str(simSummaryID) + '.csv'
    print("Reading sim summary from " + allSimsDir + oldSimSummaryFile)
    simSummary = pd.read_csv(allSimsDir + oldSimSummaryFile)
    # If simSummary contains 'Run' and 'job' columns, remove them.
    # Later, the index will be used to create a new 'Run' column.
    try:
        simSummary.drop(['Run', 'job'], inplace=True)
    except ValueError as e:
        pass
    simSummary.drop_duplicates(inplace=True)


    if useNewJobID:
        # Create a unique ID for the simulations (concatenation of year, month, day, hour, minute, second).
        newJobID = (str(datetime.datetime.today()).split('.')[0]).replace("-", "").replace(":", "").replace(" ", "")
        print('Creating sims with ID ' + str(newJobID))
        # ID should be unique, so check whether there is evidence it has been used before.
        assert (len(glob.glob(allSimsDir + '*' + newJobID + '*')) == 0), "Found filenames and/or directory names in " + allSimsDir + " containing ID " + newJobID + ". ID must be unique."
        # Log a one-line summary.
        with open('HLBMakeJobs.' + platform.node() + '.log', 'a') as logFile:
            logFile.write(newJobID + ' ' + str(len(simSummary)) + ' (Copy of ' + str(simSummaryID) + ') ' + message + '\n')
        newSimsDir = allSimsDir + 'ID_' + newJobID + '/'
        os.mkdir(newSimsDir)
    else:
        newJobID = simSummaryID
        newSimsDir = allSimsDir + 'ID_' + newJobID + '/'

    # Declare variables that will be used for all runs.
    metaparams = dict()
    metaparams['season'] = 3 # Different numbers correspond to different types of trees, affecting yield. Bassanezi 2011.
    metaparams['profitabilityThreshold'] = 0.6
    metaparams['yieldVariationParam'] = 0.5
    # Record these variables in a file.
    with open(newSimsDir + 'metaparams_ID_' + newJobID, 'w') as metaparamsFile:
        for thisKey in metaparams.keys():
            metaparamsFile.write(thisKey + ' ' + str(metaparams[thisKey]) + '\n')

    # Create a directory (with necessary contents) for each job, and update summary file.
    # Each job makes one or more calls to the model exe, with the total running time per job intended to be around an hour.
    # The number of sims taking around an hour is given by simsPerJob.
    # The code below adds as many rows of simSummary to a job as it can, until the number of sims would exceed simsPerJob.
    # Each job has at least one row of simSummary, so simsPerJob may be exceeded if the first row of the job uses a lot of sims.
    asInt = "{:.0f}".format # Function to write integers without a decimal component.
    simsPerJob = 1000 # Number of sims that should take approximately 1 hour to run.
    MBPerJob = 20 # Size of HLBAppsAndInput.
    noMoreRowsFlag = False
    job = 0
    row = 0
    rowsInThisJob = 0
    simsInThisRow = simSummary['nParamSets'].iloc[row] * abs(simSummary['nSimsPerParamSet'].iloc[row])
    while row < len(simSummary):
        simsInThisJob = 0
        rowInThisJob = 0
        # Create directory with exes and input files.
        runDir = newSimsDir + '/ID_' + newJobID + '_Job_' + str(job)
        shutil.copytree(inputDir, runDir, ignore=shutil.ignore_patterns('*.ipdb','*.iobj','*.pdb','.svn')) # Ignore some large files that are created by the Visual Studio compiler, and SVN files.
        # Create Python script to be run for that directory.
        with open(runDir + '/runMe.py', 'w') as executableFile:
            executableFile.write('#!/usr/bin/python3\n')
            executableFile.write('# This file was automatically generated by HLBMakeJobs.py at date/time ' + str(datetime.datetime.today()) + '.\n')
            executableFile.write('import hlbutils\n')
            executableFile.write('import time\n')
            executableFile.write("print('Starting runMe...')\n")
            executableFile.write("startTime = time.time()\n")
            while (row < len(simSummary)) and ((rowInThisJob == 0) or (simsInThisJob + simsInThisRow <= simsPerJob)):
                if (row % 500) == 0 and row > 0:
                    print('Creating job files for row ' + str(row) + ' of ' + str(len(simSummary)) + ' in simSummary...')
                    print('Currently creating job ' + str(job) + '. Have probably created around ' + str(job*MBPerJob) + 'MB of job files so far.')
                simsInThisJob = simsInThisJob + simsInThisRow
                executableFile.write('exeArgs = hlbutils.runSims(nParamSets=' + asInt(simSummary['nParamSets'].iloc[row]) + \
                ', nSimsPerParamSet=' + asInt(simSummary['nSimsPerParamSet'].iloc[row]) + \
                ', surveysPerYear=' + asInt(simSummary['surveysPerYear'].iloc[row]) + \
                ", MCparamsFile='" + simSummary['MCparamsFile'].iloc[row] + \
                "', controlBetaPercent=" + str(simSummary['controlBeta'].iloc[row]) + \
                ', controlEpsilonPercent=' + str(simSummary['controlEpsilon'].iloc[row]) + \
                ', cullRadius=' + asInt(simSummary['cullRadius'].iloc[row]) + \
                ', percentageOfSymptomaticsDetected=' + asInt(simSummary['percentageOfSymptomaticsDetected'].iloc[row]) + \
                ", outputDir='', RTreesContributeToYield=" + asInt(simSummary['RTreesContributeToYield'].iloc[row]) + \
                ", surveyStartPrevPercent=" + asInt(simSummary['surveyStartPrevPercent'].iloc[row]) + \
                ", nGroups=" + asInt(simSummary['nGroups'].iloc[row]) + \
                ", controlBeta1Percent=" + str(simSummary['controlBeta1'].iloc[row]) + \
                ', controlEpsilon1Percent=' + str(simSummary['controlEpsilon1'].iloc[row]) + \
                ', row=' + asInt(row) + \
                ", locationFile='" + simSummary['locationFile'].iloc[row] + "'" + \
                ", crypticPeriodScalingFactor=" + asInt(simSummary['crypticPeriodScalingFactor'].iloc[row]) + \
                ", exposedPeriodScalingFactor=" + asInt(simSummary['exposedPeriodScalingFactor'].iloc[row]) + \
                ", exposedPeriodDistribution=" + asInt(simSummary['exposedPeriodDistribution'].iloc[row]) + \
                ", plantYear=" + asInt(simSummary['plantYear'].iloc[row]) + \
                ", simYearsBase=" + asInt(simSummary['simDurationYears'].iloc[row]) + \
                ", makeChangeTimeOutputFile=" + asInt(simSummary['makeChangeTimeOutputFile'].iloc[row]) + \
                ", yieldVariationParam=" + str(metaparams['yieldVariationParam']) + \
                ", primaryInfectionType=" + asInt(simSummary['primaryInfectionType'].iloc[row]) + \
                ", primaryInfectionDistanceParameter=" + asInt(simSummary['primaryInfectionDistanceParameter'].iloc[row]) + \
                ", crypticPeriodType=" + asInt(simSummary['crypticPeriodType'].iloc[row]) + \
                ")\n")
                simSummary.loc[row, 'job'] = job
                rowInThisJob = rowInThisJob + 1
                row = row + 1
                if row < len(simSummary):
                    simsInThisRow = simSummary['nParamSets'].iloc[row] * abs(simSummary['nSimsPerParamSet'].iloc[row])
            executableFile.write("endTime = time.time()\n")
            executableFile.write("print('Finished runMe. Took ' + str((endTime - startTime)/3600) + ' hours.')\n")
        # Add hlbutils to that directory.
        shutil.copyfile('hlbutils.py', runDir + '/hlbutils.py')
        job = job + 1

    # Generate wrapper that can be called by Condor.
    with open(newSimsDir + 'condor_wrapper_ID_' + newJobID + '.py', 'w') as condorWrapper:
        condorWrapper.write('#!/usr/bin/python3\n')
        condorWrapper.write('# This file was automatically generated by HLBMakeJobs.py\n')
        condorWrapper.write('import runMe\n')

    # Generate wrapper that can be run locally.
    with open(newSimsDir + 'local_wrapper_ID_' + newJobID + '.py', 'w') as localWrapper:
        localWrapper.write('# This file was automatically generated by HLBMakeJobs.py\n')
        localWrapper.write("import os\n")
        localWrapper.write("cwd = os.getcwd()\n")
        localWrapper.write("for iJob in range(" + str(job) + "):\n")
        localWrapper.write("    print('Starting job ' + str(iJob) + '...')\n")
        localWrapper.write("    os.chdir(cwd + '/ID_" + newJobID + "_Job_' + str(iJob) + '/')\n")
        localWrapper.write("    import runMe\n")
        localWrapper.write("    reload(runMe)\n")
        localWrapper.write("    print('Finished job ' + str(iJob) + '.')\n")
        localWrapper.write("    os.chdir(cwd)\n")

    # Generate condor script.
    with open(newSimsDir + 'condor' + '_ID_' + newJobID + '.sub', 'w') as condorSub:
        condorSub.write('  ########################\n')
        condorSub.write('  # Submit file for huanglongbing model\n')
        condorSub.write('  ########################\n')
        condorSub.write('  Executable     = condor_wrapper_ID_' + newJobID + '.py\n')
        condorSub.write('  Universe       = vanilla\n')
        condorSub.write('  arguments      = $(Process)\n')
        condorSub.write('  Output         = hlb.out.$(Process)\n')
        condorSub.write('  Log            = hlb.log.$(Process)\n')
        condorSub.write('  Error          = hlb.err.$(Process)\n')
        condorSub.write('  initialDir     = ID_' + newJobID + '_Job_$(Process)\n')
        condorSub.write('  transferInputFiles = .\n')
        condorSub.write('\n')
        condorSub.write('Queue ' + str(job))

    # Create a summary file that shows what parameter values are used in each run.
    # Make sure columns have correct dtype, so that they are saved with the correct decimal places.
    for intVar in ['nParamSets', 'nSimsPerParamSet', 'surveysPerYear', 'cullRadius', 'nGroups', 'plantYear', 'simDurationYears', 'surveyStartPrevPercent', 'RTreesContributeToYield', 'job', 'crypticPeriodScalingFactor', 'exposedPeriodScalingFactor', 'exposedPeriodDistribution', 'percentageOfSymptomaticsDetected', 'makeChangeTimeOutputFile', 'primaryInfectionType', 'crypticPeriodType']:
        simSummary[intVar] = simSummary[intVar].astype('int32')
    simSummary.to_csv(newSimsDir + 'summary_ID_' + newJobID + '.csv', header=True, index_label='Run')

    # Print some output.
    totalSims = sum(simSummary['nParamSets'] * abs(simSummary['nSimsPerParamSet']))
    print('Generated files for ' + str(len(simSummary)) + ' run(s) containing ' + str(totalSims) + ' simulations, split over ' + str(job) + ' job(s).')
    #print('Should take approximately ' + str(1.0 * totalSims / simsPerJob / nAshCores) + ' hour(s) on Condor (using ' + str(nAshCores) + ' Ash cores), or ' + str(1.0 * totalSims / simsPerJob) + ' hour(s) locally.')
    print('Simulation ID: ' + newJobID)
    print('Run on Condor from ' + newSimsDir + ' with:\n  condor_submit condor_ID_' + newJobID + '.sub')
    print('Run locally from ' + newSimsDir + ' with:\n  python local_wrapper_ID_' + newJobID + '.py')

if __name__ == "__main__":
    if len(sys.argv) > 2:
        message = sys.argv[2]
    else:
        message = ""
    makeJobsFromSimSummary(simSummaryID=sys.argv[1], message=message, useNewJobID=True)
