# @authors: Andrew P. Craig
# @email: apc53@cam.ac.uk
# @date: 2017/10/09
# @copyright: MIT License

import matplotlib
import hlbutils
import hlbplot
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import hlbplot
import pdb
import pandas as pd
import os
import glob
import sys
import subprocess
from collections import defaultdict
import itertools

# Set ID of simulation set for which figures are to be generated.
runID = sys.argv[1]
if len(sys.argv) > 2:
    figToStartFrom = int(sys.argv[2])
else:
    figToStartFrom = 0
if len(sys.argv) > 3:
    figToFinishAt = int(sys.argv[3])
else:
    figToFinishAt = np.inf
print('Generating figures ' + str(figToStartFrom) + ' - ' + str(figToFinishAt) + ' for ID ' + runID + '...')

simsDir = '../HLBSims/ID_' + runID + '/'
figsPrefix = '../HLBFigs/' + runID + '_'
inputDir = '../HLBAppsAndInput/'

# Load metaparameters (season etc.).
intMetaparams = ['nParamSets', 'season', 'nSimsPerParamSet', 'Tinitial']
floatMetaparams = ['profitabilityThreshold', 'yieldVariationParam']
metaparams = dict()
with open(simsDir + 'metaparams_ID_' + runID, 'r') as metaparamsFile:
    for line in metaparamsFile.read().splitlines():
        thisKey, thisValue = tuple(line.split(' '))
        if thisKey in intMetaparams:
            metaparams[thisKey] = int(thisValue)
        elif thisKey in floatMetaparams:
            metaparams[thisKey] = float(thisValue)
        else:
            metaparams[thisKey] = thisValue

# Load simulation summary file.
simSummary = pd.read_csv(simsDir + 'summary_ID_' + runID + '.csv')

simSummaryReduced = simSummary.copy()[['MCparamsFile', 'locationFile', 'nParamSets', 'plantYear', 'simDurationYears']]
simSummaryReduced = simSummaryReduced.drop_duplicates()
# Create dictionaries that allow MCparams file names, plant years etc. to be looked up by subregion ID.
MCparamsFileList = simSummaryReduced['MCparamsFile'].tolist()
subregionID = ['Young', 'Old']
MCparamsFile =     dict(zip(subregionID, MCparamsFileList))
locationFile =     dict(zip(subregionID, simSummaryReduced['locationFile'].tolist()))
nParamSets =       dict(zip(subregionID, simSummaryReduced['nParamSets'].tolist()))
nSubregions =      dict(zip(subregionID, [1]*len(MCparamsFile)))
plantYear =        dict(zip(subregionID, simSummaryReduced['plantYear'].tolist()))
simDurationYears = dict(zip(subregionID, simSummaryReduced['simDurationYears'].tolist()))

controlBeta = {'NoControl':0, 'Control':90, 'SprayingOnly':90, 'RoguingOnly':0, 'BetaReductionOnly':95, 'EpsilonReductionOnly':0, 'PerfectDetectionOnly':0, 'NoControlPerfectDetection':0, 'ControlPerfectDetection':90}
controlEpsilon = {'NoControl':0, 'Control':90, 'SprayingOnly':90, 'RoguingOnly':0, 'BetaReductionOnly':0, 'EpsilonReductionOnly':95, 'PerfectDetectionOnly':0, 'NoControlPerfectDetection':0, 'ControlPerfectDetection':90}
surveysPerYear = {'NoControl':0, 'Control':4, 'SprayingOnly':0, 'RoguingOnly':4, 'BetaReductionOnly':4, 'EpsilonReductionOnly':4, 'PerfectDetectionOnly':4, 'NoControlPerfectDetection':0, 'ControlPerfectDetection':4}
percentageOfSymptomaticsDetected = {'NoControl':50, 'Control':50, 'RoguingOnly':50, 'PerfectDetectionOnly':100, 'NoControlPerfectDetection':100, 'ControlPerfectDetection':100}
title = {'NoControl':'No control', 'Control':'Control', 'SprayingOnly':'Spraying only', 'RoguingOnly':'Roguing only (50% detection)', 'BetaReductionOnly':'Secondary control (95%) only', 'EpsilonReductionOnly':'Primary control (95%) only', 'PerfectDetectionOnly':'Roguing only (100% detection)', 'NoControlPerfectDetection':'No control', 'ControlPerfectDetection':'Control'}
subtitle = defaultdict(lambda: '', {'Control':str(surveysPerYear['Control']) + ' surveys per year\n' + str(controlBeta['Control']) + '% reduction in infection rates'})

# Set parameters for simulations showing sensitivity to primary infection at edge only, and short latent period.
# Values from fitting.
oldShortLatentControlBeta = 79
oldEdgeShortLatentControlBeta = 77
youngShortLatentControlBeta = 80
youngEdgeShortLatentControlBeta = 73
# Function to get controlBeta from above values, when there is also 90% control.
control90PercentFn = lambda x: (100 - x) * 0.9 + x
edgeShortLatentSensitivityParams = {}
for x in ['Old', 'Young']:
    edgeShortLatentSensitivityParams.update({ \
                            (x, 'normal', 'NoControl'):{'primaryInfectionType':0, 'exposedPeriodDistribution':1, 'crypticPeriodType':1, 'controlBeta':0, 'controlEpsilon':0, 'surveysPerYear':0},
                            (x, 'normal', 'Control'):{'primaryInfectionType':0, 'exposedPeriodDistribution':1, 'crypticPeriodType':1, 'controlBeta':90, 'controlEpsilon':90, 'surveysPerYear':4},
                            (x, 'edge', 'NoControl'):{'primaryInfectionType':1, 'exposedPeriodDistribution':1, 'crypticPeriodType':1, 'controlBeta':0, 'controlEpsilon':0, 'surveysPerYear':0},
                            (x, 'edge', 'Control'):{'primaryInfectionType':1, 'exposedPeriodDistribution':1, 'crypticPeriodType':1, 'controlBeta':90, 'controlEpsilon':90, 'surveysPerYear':4}})
edgeShortLatentSensitivityParams.update({ \
                         ('Old', 'shortLatent', 'NoControl'):{'primaryInfectionType':0, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':oldShortLatentControlBeta, 'controlEpsilon':0, 'surveysPerYear':0},
                         ('Old', 'edgeShortLatent', 'NoControl'):{'primaryInfectionType':1, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':oldEdgeShortLatentControlBeta, 'controlEpsilon':0, 'surveysPerYear':0},
                         ('Old', 'shortLatent', 'Control'):{'primaryInfectionType':0, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':control90PercentFn(oldShortLatentControlBeta), 'controlEpsilon':90, 'surveysPerYear':4},
                         ('Old', 'edgeShortLatent', 'Control'):{'primaryInfectionType':1, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':control90PercentFn(oldEdgeShortLatentControlBeta), 'controlEpsilon':90, 'surveysPerYear':4},
                         ('Young', 'shortLatent', 'NoControl'):{'primaryInfectionType':0, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':youngShortLatentControlBeta, 'controlEpsilon':0, 'surveysPerYear':0},
                         ('Young', 'edgeShortLatent', 'NoControl'):{'primaryInfectionType':1, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':youngEdgeShortLatentControlBeta, 'controlEpsilon':0, 'surveysPerYear':0},
                         ('Young', 'shortLatent', 'Control'):{'primaryInfectionType':0, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':control90PercentFn(youngShortLatentControlBeta), 'controlEpsilon':90, 'surveysPerYear':4},
                         ('Young', 'edgeShortLatent', 'Control'):{'primaryInfectionType':1, 'exposedPeriodDistribution':5, 'crypticPeriodType':3,
                                'controlBeta':control90PercentFn(youngEdgeShortLatentControlBeta), 'controlEpsilon':90, 'surveysPerYear':4}})
# Add on parameters that are same for all.
for (age, simType, control) in edgeShortLatentSensitivityParams:
    edgeShortLatentSensitivityParams[(age, simType, control)].update( \
        {'nSimsPerParamSet':1,
        'nParamSets':nParamSets[age], 'MCparamsFile':MCparamsFile[age], 'locationFile':locationFile[age], 'plantYear':plantYear[age], 'simDurationYears':simDurationYears[age],
        'percentageOfSymptomaticsDetected':50, 'cullRadius':0, 'primaryInfectionDistanceParameter':10})
# Done constructing sensitivity sim parameters.

print('About to make figures ' + str(figToStartFrom) + '-' + str(figToFinishAt) + '...')

if figToStartFrom <= 1:
    print('Trying to generate Fig. 1, but there is no code. (It is a schematic.)')

for (subregion, figsSuffix, trajectoryToHighlight) in [('Young', '_Young.png', {'sim':600, 'seed':0}), ('Old', '_Old.png', {'sim':499, 'seed':0})]:
    print('Subregion: ' + str(subregion))

    thisFig = 2
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 5
        ncols = 2
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, gridspec_kw={'height_ratios':([1] * nrows)})
        for indControl, controlLevel in enumerate(['NoControl', 'Control']):
            try:
                # No control, epidemic time course.
                for indCompartment, (thisCompartment, thisCompartmentLabel, annotationY, thisTitle, thisSubtitle) in enumerate(zip(['S', 'E', 'I', 'D', 'R'], ['S', 'E', 'C', 'I', 'R'], [40] + [80] * 4, [title[controlLevel]] + [''] * 4, [subtitle[controlLevel]] + [''] * 4)):
                    if indCompartment == 0:
                        subplotLetter = 'ab'[indControl]
                    else:
                        subplotLetter = ''
                    data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel],
                         controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel], cullRadius=0, MCparamsFile=MCparamsFile[subregion],
                         nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'],
                         asPercentages=True, heightAsProportionOfWidth=0.66, xMax=simDurationYears[subregion], ax=axarr[indCompartment, indControl], rowTitle=rowTitle, theLegends=theLegends, rowTitleY=-0.1, compartments=[thisCompartment], plotType='pointwiseQuantiles',
                         textItems=textItems, showXLabel=(indCompartment == (nrows - 1)), showYLabel=((indCompartment==2) and (controlLevel=='NoControl')), yTicks=[0, 50, 100], annotation=(thisCompartmentLabel, 0.3, annotationY), subplotLetter=subplotLetter, title=thisTitle)
            except IOError as theError:
                print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
        # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, heightScalingFactor=0.3, hspace=0.4)
        print('Saved figure: ' + figName)

    thisFig = 3
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 2
        ncols = 2
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, gridspec_kw={'height_ratios':[1,1]})
        for indControl, controlLevel in enumerate(['NoControl', 'Control']):
            try:
                data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel],
                    controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel], cullRadius=0,
                     MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion],
                     simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'], title=title[controlLevel], asPercentages=False, heightAsProportionOfWidth=0.66,
                     xMax=simDurationYears[subregion], ax=axarr[0, indControl], rowTitle=rowTitle, theLegends=theLegends, rowTitleY=-0.1, compartments=['actualPrev', 'detectablePrev'],
                     plotLabels=['Actual', 'Detectable'], plotType='pointwiseQuantiles', subplotLetter='ab'[indControl], textItems=textItems, RThreshold=(7700/1))
                if controlLevel == 'NoControl':
                    interpolateOverX = True
                    plotType = 'pointwiseQuantiles'
                else:
                    interpolateOverX = False
                    plotType = 'binnedQuantiles'
                data, d, nSims, nSeeds, nTrees = hlbplot.plotSeriesAgainstSeries(hlbutils.getOutputPath(simsDir, runID, simSummary,
                    percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel], surveysPerYear=surveysPerYear[controlLevel], controlBeta=controlBeta[controlLevel],
                    controlEpsilon=controlEpsilon[controlLevel], cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion],
                    nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'], title='',
                    asPercentages=False, heightAsProportionOfWidth=0.66, xMax=30, ax=axarr[1, indControl], rowTitle=rowTitle, rowTitleY=-0.1, yContents='actualPrev', xContents='detectablePrev',
                    profitabilityThreshold=metaparams['profitabilityThreshold'], RThreshold=(7700/1), plotType=plotType, subplotLetter='cd'[indControl], textItems=textItems,
                    actualPrevsToHighlight=[0.9, 1.0], actualPrevsAndPercentilesToSummarise=[(0.6,0.5), (0.8,0.5)], interpolateOverX=interpolateOverX, yTicks=range(0, 101, 20),
                    xTicks=range(0, 31, 5))
            except IOError as theError:
                print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
        # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems)
        print('Saved figure: ' + figName)

    thisFig = 4
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        AUCScalingDivisor = 10000
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 3
        ncols = 2
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, gridspec_kw={'height_ratios':([1] * nrows)})
        for indControl, controlLevel in enumerate(['NoControl','Control']):
            try:
                # No control, epidemic time course.
                yLabel='AUC of C+I\n(tree-years $\\times$ ' + "{:,}".format(AUCScalingDivisor) + ')'
                data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel],
                    controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel],
                    cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet,
                    plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'], title=title[controlLevel],
                    asPercentages=False, heightAsProportionOfWidth=0.66, xMax=simDurationYears[subregion], ax=axarr[0, indControl], rowTitle=rowTitle, theLegends=theLegends,
                    rowTitleY=-0.1, plotLabels=[], compartments=['I+Dcumulative'], plotType='pointwiseQuantiles', subplotLetter='abxx'[indControl], textItems=textItems, yMax=7,
                    yLabel=yLabel, yScalingDivisor=AUCScalingDivisor, greyscale=True, yTicks=range(8))
                if (subregion == 'Young') and (controlLevel == 'NoControl'):
                    profitabilityTextAtRight = True
                else:
                    profitabilityTextAtRight = False
                data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel],
                    controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel],
                    cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet,
                    plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'], title='', asPercentages=False,
                    heightAsProportionOfWidth=0.66, xMax=simDurationYears[subregion], ax=axarr[1, indControl], rowTitle=rowTitle, theLegends=theLegends, rowTitleY=-0.1,
                    compartments=['treeYieldLoss'], plotType='pointwiseQuantiles', plotPotentialYield=False, subplotLetter='cd'[indControl], textItems=textItems,
                    yLabel='Yield loss (harvests)', plotProfitabilityThreshold=metaparams['profitabilityThreshold'], profitabilityTextAtRight=profitabilityTextAtRight,
                    yTicks=np.arange(0, 1.21, 0.2))
                data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel],
                    controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel],
                    cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet,
                    plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'], title='', asPercentages=False,
                    heightAsProportionOfWidth=0.66, xMax=simDurationYears[subregion], ax=axarr[2, indControl], rowTitle=rowTitle, theLegends=theLegends, rowTitleY=-0.1,
                    compartments=['treeYieldLosscumulative'], plotType='pointwiseQuantiles', plotPotentialYield=False, subplotLetter='ef'[indControl], textItems=textItems,
                    yLabel='Cumulative yield loss \n(harvests)', yMax=6, yTicks=range(7))
            except IOError as theError:
                print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
                # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, heightScalingFactor=0.75)
        print('Saved figure: ' + figName)

    thisFig = 5
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        AUCScalingDivisor = 10000
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 2
        ncols = 1
        controlBetasEpsilons = np.linspace(0,100,num=11).round().astype(int) # expspace returns round numbers, but hlbutils.getOutputPath expects ints so convert them here.
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols)
        figFileName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        try:
            outputPathsBetaVarying = {}
            outputPathsBetaLow = {}
            outputPathsEpsilonLow = {}
            surveysPerYearAll = [0, 1, 2, 4]
            for (thisSurveysPerYear, thisControl) in itertools.product(surveysPerYearAll, controlBetasEpsilons):
                outputPathsBetaVarying[(thisSurveysPerYear, thisControl)] = hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=thisSurveysPerYear, controlBeta=thisControl, controlEpsilon=thisControl, percentageOfSymptomaticsDetected=50, cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion])
            # Column with epsilon and beta varying
            yLabel='AUC of C+I\n(tree-years $\\times$ ' + "{:,}".format(AUCScalingDivisor) + ')'
            hlbplot.plotIndicatorByVarying(outputPathsBetaVarying, figFileName, surveysPerYear=surveysPerYearAll, indicatorCompartment='I+Dcumulative', indicatorSummaryFn='valAtHarvest', xContents=controlBetasEpsilons, season=metaparams['season'], profitabilityThreshold=metaparams['profitabilityThreshold'], ax=axarr[0], RThreshold=7700, plotType='boxplot', textItems=textItems, yLog=False, subplotLetter='a', yLabel=yLabel, yMin=0, yMax=7, yTicks=range(7+1), yScalingDivisor=AUCScalingDivisor)
            hlbplot.plotIndicatorByVarying(outputPathsBetaVarying, figFileName, surveysPerYear=surveysPerYearAll, indicatorCompartment='treeYieldLosscumulative', indicatorSummaryFn='valAtHarvest', xContents=controlBetasEpsilons, season=metaparams['season'], profitabilityThreshold=metaparams['profitabilityThreshold'], ax=axarr[1], RThreshold=7700, plotType='boxplot', textItems=textItems, subplotLetter='b', yMin=0, yMax=9, yTicks=range(9+1), yLabel='Cumulative yield loss\n(harvests)')
        except IOError as theError:
            print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
        # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        print('textItems: ' + str(textItems))
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, widthScalingFactor=2)
        print('Saved figure: ' + figName)

    thisFig = 6
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        AUCScalingDivisor = 10000
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 2
        ncols = 2
        detectionEfficiencies = np.linspace(50,100,num=6)
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols)
        figFileName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        try:
            for indControl, controlLevel in enumerate(['NoControl', 'Control']):
                outputPaths = {}
                surveysPerYearAll = [0, 1, 2, 4]
                for (thisSurveysPerYear, thisDetectionEfficiency) in itertools.product(surveysPerYearAll, detectionEfficiencies):
                    outputPaths[(thisSurveysPerYear,thisDetectionEfficiency)] = hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=thisSurveysPerYear, controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=thisDetectionEfficiency, cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion])
                # Column with epsilon and beta varying
                yLabel='AUC of C+I\n(tree-years $\\times$ ' + "{:,}".format(AUCScalingDivisor) + ')'
                hlbplot.plotIndicatorByVarying(outputPaths, figFileName, surveysPerYear=surveysPerYearAll, indicatorCompartment='I+Dcumulative', indicatorSummaryFn='valAtHarvest', xContents=detectionEfficiencies, season=metaparams['season'], profitabilityThreshold=metaparams['profitabilityThreshold'], ax=axarr[0,indControl], RThreshold=7700, plotType='boxplot', textItems=textItems, yLog=False, subplotLetter='ab'[indControl], yLabel=yLabel, yMin=0, yMax=7, yTicks=range(7+1), yScalingDivisor=AUCScalingDivisor, xMin=50, title=title[controlLevel], xLabel='Detection efficiency (%)', nCols=ncols, subplotLetterXOffset=0.2)
                hlbplot.plotIndicatorByVarying(outputPaths, figFileName, surveysPerYear=surveysPerYearAll, indicatorCompartment='treeYieldLosscumulative', indicatorSummaryFn='valAtHarvest', xContents=detectionEfficiencies, season=metaparams['season'], profitabilityThreshold=metaparams['profitabilityThreshold'], ax=axarr[1,indControl], RThreshold=7700, plotType='boxplot', textItems=textItems, yLog=False, subplotLetter='cd'[indControl], yMin=0, yMax=7, yTicks=range(7+1), yScalingDivisor=AUCScalingDivisor, xMin=50, yLabel='Cumulative yield loss\n(harvests)', xLabel='Detection efficiency (%)', nCols=ncols, subplotLetterXOffset=0.2)
        except IOError as theError:
            print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
        # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        print('textItems: ' + str(textItems))
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, widthScalingFactor=1)
        print('Saved figure: ' + figName)

    thisFig = 8
    if figToStartFrom <= thisFig <= figToFinishAt:
        # Had a lot of trouble getting time course plot in first row, followed by 2x2 grid of spatial plots,
        # so making them as separate images, which are then stitched together with imagemagick.
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        nSimsPerParamSet = 1
        rowTitle = ''
        # Create time course plot.
        theLegends = []
        textItems = []
        nrows = 1
        ncols = 1
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols)
        for indControl, controlLevel in enumerate(['NoControl']):
            try:
                data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel], controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel], cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion], makeChangeTimeOutputFile=1), ax=axarr, figFileName='', season=metaparams['season'], title='', asPercentages=True, xMax=simDurationYears[subregion], rowTitle=rowTitle, theLegends=theLegends, rowTitleY=-0.1, compartments=['S', 'E', 'I', 'D', 'R'], plotLabels=['S', 'E', 'C', 'I', 'R'], plotType='pointwiseQuantiles', subplotLetter='a'[indControl], textItems=textItems, RThreshold=(7700/1), onlyTrajectoryToHighlight=True, trajectoryToHighlight=trajectoryToHighlight, subplotLetterXOffset=0.1)
            except IOError as theError:
                print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
                # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + '_1' + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=2, textItems=textItems, heightScalingFactor=0.6)
        # Reset variables and create 2x2 grid of spatial plots.
        theLegends = []
        textItems = []
        nrows = 2
        ncols = 2
        rotation = 0 # Use 0.025 for original data to rotate them so that they look straighter.
        resultsFileColNames = ["block", "c", "x", "y", "dmin", "dmax", "rmv", "plantYear", "sim", "tree", "E", "I", "D", "R", "Tinitial", "seed", "group"]
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, sharey=True, gridspec_kw={'height_ratios':[1,1]})
        for indControl, controlLevel in enumerate(['NoControl']):
            for (frameTime, axarrIndex, subplotLetter, showScale) in zip(np.arange(0.5, 2.1, 0.5), [(0,0), (0,1), (1,0), (1,1)], 'bcde', [False, False, False, True]):
                try:
                    # Get results file name, dropping the '0' at the end because the time-change file doesn't have that.
                    resultsFileName = hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel], controlBeta=controlBeta[controlLevel],
                        controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel], cullRadius=0, MCparamsFile=MCparamsFile[subregion],
                        nParamSets=nParamSets[subregion], locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion],
                        makeChangeTimeOutputFile=1)[0:-1]
                    # These files can be very large, and we only need the results for one sim/seed combination, so extract that using awk if it doesn't already exist.
                    resultsFileNameJustTrajectorySimSeed = resultsFileName + '_Sim' + str(trajectoryToHighlight['sim']) + 'Seed' + str(trajectoryToHighlight['seed'])
                    if not os.path.isfile(resultsFileNameJustTrajectorySimSeed):
                        awkCall = "awk '$" + str(resultsFileColNames.index('sim') + 1) + " == " + str(trajectoryToHighlight['sim']) + ", $" + str(resultsFileColNames.index('seed') + 1) + " == " + str(trajectoryToHighlight['seed']) + "' " + resultsFileName + " > " + resultsFileNameJustTrajectorySimSeed
                        print("Extracting just necessary sim/seed with following call: ")
                        print(awkCall)
                        subprocess.check_call(awkCall, shell=True)
                    print("Loading file: " + resultsFileNameJustTrajectorySimSeed)
                    data = pd.read_csv(resultsFileNameJustTrajectorySimSeed, delimiter = "\s", index_col=None, names=resultsFileColNames)
                    hlbplot.plotHLBGridOverTime(data, frameTimes=[frameTime], sim=trajectoryToHighlight['sim'], seed=trajectoryToHighlight['seed'], ax=axarr[axarrIndex], subplotLetter=subplotLetter,
                        textItems=textItems, showScale=showScale, rotation=rotation)
                except IOError as theError:
                    print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
            # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + '_2' + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, heightScalingFactor=0.8, hspace=0.1)
        # Stitch both images together with imagemagick.
        inputFigNames = figsPrefix + 'Fig' + str(thisFig) + '_*' + figsSuffix
        outputFigName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        subprocess.check_call("convert -append " + inputFigNames + ' ' + outputFigName, shell=True)
        print('Saved figure: ' + outputFigName)


    thisFig = 9
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 2
        ncols = 2
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, gridspec_kw={'height_ratios':[1,1]})
        for indControl, controlLevel in enumerate(['NoControlPerfectDetection','ControlPerfectDetection']):
            try:
                # No control, epidemic time course.
                data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=hlbutils.getOutputPath(simsDir, runID, simSummary, surveysPerYear=surveysPerYear[controlLevel], controlBeta=controlBeta[controlLevel],
                    controlEpsilon=controlEpsilon[controlLevel], percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel], cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion],
                    locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'],
                    title=title[controlLevel], asPercentages=False, heightAsProportionOfWidth=0.66, xMax=simDurationYears[subregion], ax=axarr[0, indControl], rowTitle=rowTitle, theLegends=theLegends, rowTitleY=-0.1,
                    compartments=['actualPrev', 'detectablePrev'], plotLabels=['Actual', 'Detectable'], plotType='pointwiseQuantiles', subplotLetter='ab'[indControl], textItems=textItems, RThreshold=(7700/1))
                if controlLevel == 'NoControlPerfectDetection':
                    interpolateOverX = True
                    plotType = 'pointwiseQuantiles'
                else:
                    interpolateOverX = False
                    plotType = 'binnedQuantiles'
                data, d, nSims, nSeeds, nTrees = hlbplot.plotSeriesAgainstSeries(hlbutils.getOutputPath(simsDir, runID, simSummary, percentageOfSymptomaticsDetected=percentageOfSymptomaticsDetected[controlLevel],
                    surveysPerYear=surveysPerYear[controlLevel], controlBeta=controlBeta[controlLevel], controlEpsilon=controlEpsilon[controlLevel], cullRadius=0, MCparamsFile=MCparamsFile[subregion], nParamSets=nParamSets[subregion],
                    locationFile=locationFile[subregion], nSimsPerParamSet=nSimsPerParamSet, plantYear=plantYear[subregion], simDurationYears=simDurationYears[subregion]), figFileName='', season=metaparams['season'], title='',
                    asPercentages=False, heightAsProportionOfWidth=0.66, xMax=50, ax=axarr[1, indControl], rowTitle=rowTitle, rowTitleY=-0.1, yContents='actualPrev', xContents='detectablePrev',
                    profitabilityThreshold=metaparams['profitabilityThreshold'], RThreshold=(7700/1), plotType=plotType, subplotLetter='cd'[indControl], textItems=textItems,
                    actualPrevsAndPercentilesToSummarise=[(0.6,0.5), (0.8,0.5)], interpolateOverX=interpolateOverX, xTicks=range(0, 51, 10))
            except IOError as theError:
                print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
                # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems)
        print('Saved figure: ' + figName)

    thisFig = 10
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        nrows = 4
        ncols = 2
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, sharey=True, sharex=True, gridspec_kw={'height_ratios':[1] * nrows})
        try:
            for indRow, (sensitivityVersion, rowTitle) in enumerate(zip(['normal', 'edge', 'shortLatent', 'edgeShortLatent'], ['Baseline\n', 'Edge\ninfection', 'Short\nlatent', 'Edge infection,\nshort latent'])):
                for indCol, controlLevel in enumerate(['NoControl', 'Control']):
                    if controlLevel == 'NoControl':
                        interpolateOverX = True
                        plotType = 'pointwiseQuantiles'
                        if indRow == 0:
                            thisTitle = 'No control'
                            subplotLetter = 'a'
                            subplotLetterXOffset = 0.30
                        else:
                            thisTitle = ''
                            subplotLetter = ''
                    else:
                        interpolateOverX = False
                        plotType = 'binnedQuantiles'
                        if indRow == 0:
                            thisTitle = 'Control'
                            subplotLetter = 'b'
                            subplotLetterXOffset = 0.05
                        else:
                            thisTitle = ''
                            subplotLetter = ''
                    if indCol != 0:
                        rowTitle = ''
                    thisResultsParams = edgeShortLatentSensitivityParams[(subregion, sensitivityVersion, controlLevel)]
                    # Alignment does not work correctly for rotated text, so calculate adjustment for row titles based on number of characters relative to "Baseline" (which is aligned correctly).
                    rowTitleY = 0.5 + (max(map(len, rowTitle.split("\n"))) - len("Baseline")) * 0.04
                    data, d, nSims, nSeeds, nTrees = hlbplot.plotSeriesAgainstSeries(hlbutils.getOutputPath(simsDir, runID, simSummary, **thisResultsParams), figFileName='', season=metaparams['season'], title=thisTitle, asPercentages=False,
                        heightAsProportionOfWidth=0.66, xMax=30, ax=axarr[indRow, indCol], rowTitle=rowTitle, rowTitleY=rowTitleY, yContents='actualPrev', xContents='detectablePrev', profitabilityThreshold=metaparams['profitabilityThreshold'], RThreshold=(7700/2),
                        plotType=plotType, subplotLetter=subplotLetter, textItems=textItems, interpolateOverX=interpolateOverX, prev28LabelRotation=0, prev28LabelX=6, prev28LabelY=20,
                        showXLabel=(indRow == nrows-1), showYLabel=(indRow == 1 and indCol == 0), xTicks=range(0, 31, 5), subplotLetterXOffset=subplotLetterXOffset)
        except IOError as theError:
            print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
            # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, heightScalingFactor=0.5, hspace=0.4)
        print('Saved figure: ' + figName)

    thisFig = 11
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        nSimsPerParamSet = 1
        rowTitle = ''
        theLegends = []
        textItems = []
        nrows = 4
        ncols = 4
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, gridspec_kw={'height_ratios':([1] * nrows)})
        for indSimType, (thisSimType, colTitle) in enumerate(zip(['normal', 'edge', 'shortLatent', 'edgeShortLatent'], ['Baseline\n', 'Edge infection\n', 'Short latent\n', 'Edge infection,\nshort latent'])):
            try:
                thisResultsParams = edgeShortLatentSensitivityParams[(subregion, thisSimType, 'NoControl')]
                outputPath = hlbutils.getOutputPath(simsDir, runID, simSummary, **thisResultsParams)
                # No control, epidemic time course.
                for indCompartment, (thisCompartment, thisCompartmentLabel, annotationY) in enumerate(zip(['S', 'E', 'I', 'D'], ['S', 'E', 'C', 'I'], [40] + [80] * 4)):
                    if indCompartment == 0:
                        subplotLetter = 'abcd'[indSimType]
                        thisTitle = colTitle
                    else:
                        subplotLetter = ''
                        thisTitle = ''
                    if thisCompartment == 'D':
                        # We will place a marker at the point that was used for calibration (the mean (over simulations) of the timeSinceSimStarted at which D reached 50%).
                        data, _, _, _, _, _ = hlbutils.loadResults(outputPath, metaparams['season'], useYieldResultsFile=False)
                        D50BySim = data.groupby('sim').apply(hlbutils.D50)
                        markerAtPoint = {'x':D50BySim.mean(), 'y':7700/2}
                    else:
                        markerAtPoint = None
                    if indSimType == 0:
                        annotation = (thisCompartmentLabel, 0.3, annotationY)
                    else:
                        annotation = None
                    data, d, nSims, nSeeds, nTrees = hlbplot.plotTimeCourse(resultsFileName=outputPath, figFileName='', season=metaparams['season'], asPercentages=True, heightAsProportionOfWidth=0.66,
                        xMax=5, ax=axarr[indCompartment, indSimType], theLegends=theLegends, compartments=[thisCompartment], plotType='pointwiseQuantiles', textItems=textItems,
                        showXLabel=(indCompartment == (nrows - 1)), showYLabel=((indCompartment==1) and (indSimType==0)), yTicks=[0, 50, 100], annotation=annotation, subplotLetter=subplotLetter,
                        title=thisTitle, markerAtPoint=markerAtPoint, subplotLetterYOffset=0.5,
                        showYTickLabels=(indSimType==0))
            except IOError as theError:
                print('  For Fig. ' + str(thisFig) + ', required sims not found. Probably they were not run. Skipping.')
                # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, widthScalingFactor=0.5, heightScalingFactor=0.3, hspace=0.4, w_pad=-1, wspace=0.2)
        print('Saved figure: ' + figName)

    thisFig = 12
    if figToStartFrom <= thisFig <= figToFinishAt:
        print('Making Fig. ' + str(thisFig) + figsSuffix + '...')
        AUCScalingDivisor = 10000
        nSimsPerParamSet = 1
        theLegends = []
        textItems = []
        rowTitle = ''
        nrows = 2
        ncols = 2
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, sharex=True)
        figFileName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        for indCol, controlLevel in enumerate(['NoControl', 'Control']):
            outputPaths = {}
            for thisSimType in ['normal', 'edge', 'shortLatent', 'edgeShortLatent']:
                thisResultsParams = edgeShortLatentSensitivityParams[(subregion, thisSimType, 'NoControl')]
                outputPaths[thisSimType] = hlbutils.getOutputPath(simsDir, runID, simSummary, **thisResultsParams)
            yLabel='AUC of C+I\n(tree-years $\\times$ ' + "{:,}".format(AUCScalingDivisor) + ')'
            hlbplot.plotSimTypeBoxplot(outputPaths, figFileName, indicatorCompartment='I+Dcumulative', indicatorSummaryFn='valAtHarvest', xContents=['normal', 'edge', 'shortLatent', 'edgeShortLatent'],
                season=metaparams['season'], profitabilityThreshold=metaparams['profitabilityThreshold'],
                ax=axarr[0,indCol], RThreshold=7700, plotType='boxplot', textItems=textItems, yLog=False, subplotLetter='ab'[indCol], yLabel=yLabel, yMin=0, yMax=8, yTicks=range(8+1), yScalingDivisor=AUCScalingDivisor, xMin=50,
                title=title[controlLevel], xLabel='', nCols=ncols, subplotLetterXOffset=0.2)
            hlbplot.plotSimTypeBoxplot(outputPaths, figFileName, indicatorCompartment='treeYieldLosscumulative', indicatorSummaryFn='valAtHarvest', xContents=['normal', 'edge', 'shortLatent', 'edgeShortLatent'],
                season=metaparams['season'], profitabilityThreshold=metaparams['profitabilityThreshold'],
                ax=axarr[1,indCol], RThreshold=7700, plotType='boxplot', textItems=textItems, yLog=False, subplotLetter='cd'[indCol], yMin=0, yMax=7, yTicks=range(7+1), yScalingDivisor=AUCScalingDivisor, xMin=50,
                yLabel='Cumulative yield loss\n(harvests)', xLabel='', nCols=ncols, subplotLetterXOffset=0.2, xTicks=['Baseline\n', 'Edge infection\n', 'Short latent\n', 'Edge infection,\nshort latent'])
        # Save.
        figName = figsPrefix + 'Fig' + str(thisFig) + figsSuffix
        print('textItems: ' + str(textItems))
        hlbplot.tryToSaveFig(figName, theLegends=theLegends, nrows=nrows, ncols=ncols, textItems=textItems, widthScalingFactor=1)
        print('Saved figure: ' + figName)
