# @authors: Andrew P. Craig
# @email: apc53@cam.ac.uk
# @date: 2017/10/09
# @copyright: MIT License

# Plot functions.
# There is some unclear use of 'sim' to mean 'parameter set', and 'seed' to mean 'iteration'.

import os
import pickle
import hlbutils
import itertools
import matplotlib.pyplot as plt
from scipy.spatial import distance
from matplotlib.ticker import FormatStrFormatter
from numpy.testing import assert_almost_equal
from matplotlib import cm
from matplotlib import lines
from matplotlib import colors
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.cm import get_cmap
import numpy as np
import scipy
import glob
import pandas as pd
import math
import time
import stochPlot
from collections import defaultdict

fontName = 'Helvetica'

fontSizePublication = 8
fontSizeTitle = 10
fontSizePresentation = 20
fontSizeSubplotLetter = 16

lowerQuantile = 0.025
lowerMidQuantile = 0.25
midQuantile = 0.5
upperMidQuantile = 1 - lowerMidQuantile
upperQuantile = 1 - lowerQuantile
innerRangeAlpha = 0.5
outerRangeAlpha = 0.25

fill_betweenParams = dict(linewidth=0)

colours = defaultdict(lambda: 'black', **{'S':'g', 'E':'r', 'I':'b', 'D':'m', 'R':'c', 'E+I+D':'purple', 'I+D':'brown', 'treeYield':'k', 'treeYieldcumulative':'k', 'probabilityOfEscape:0.001':'violet',
                                            'probabilityOfEscape:0.0001':'purple', 'actualPrev':'goldenrod', 'detectablePrev':'orangered', 'detectablePrevInclR':'indigo', 'inoculumReleasedByYield':'k',
                                            'potentialYield':'g', 'I+Dcumulative':'brown', 'E+I+Dcumulative':'purple', '0':'green', '1':'red', '2':'blue', '4':'m', '12':'c'})

tickLength=2.5

def timeMe_decorator(f):
    def _inner(*args, **kwargs):
        startTime = time.time()
        result = f(*args, **kwargs)
        print('Took ' + "{:.1f}".format((time.time() - startTime) / 60) + ' minutes.')
        return result
    return _inner

def mean(A):
    """Returns mean of list.
    """
    return sum(A) / len(A)

def tryToSaveFig(figFileName, ncols=1, nrows=1, theLegends=None, widthScalingFactor=1, heightScalingFactor=1, textItems=None, hspace=None, wspace=None, h_pad=3, w_pad=None):
    defaultPerColWidthInMM = 65
    defaultPerRowHeightInMM = 1 * defaultPerColWidthInMM
    widthInMM = ncols * defaultPerColWidthInMM * widthScalingFactor
    heightInMM = nrows * defaultPerRowHeightInMM * heightScalingFactor
    #heightInMM = widthInMM * heightAsProportionOfWidth
    plt.gcf().set_size_inches(widthInMM/25.4, heightInMM/25.4) # Divide by 25.4 to convert mm to inches.
    figFormat = 'png'
    theDPI = 300
    plt.tight_layout(h_pad=h_pad, w_pad=w_pad)
    if hspace != None:
        plt.subplots_adjust(hspace=hspace)
    if wspace != None:
        plt.subplots_adjust(wspace=wspace)
    if theLegends != None or textItems != None:
        plt.savefig(figFileName, format=figFormat, dpi=theDPI, bbox_extra_artists=(textItems + theLegends), bbox_inches='tight')
    else:
        plt.savefig(figFileName, format=figFormat, dpi=theDPI)
    print('Saved ' + figFileName)

def plotTest():
    """Plot some random values to test tryToSaveFig
    """
    fig, axarr = plt.subplots(nrows=2, ncols=2, sharex=True, gridspec_kw={'height_ratios':([1] * 2)})
    x = np.arange(0, 10)
    axarr[0, 0].fill_between(x, x * 10, x * 40, alpha=outerRangeAlpha, **fill_betweenParams)
    axarr[0, 0].fill_between(x, x * 20, x * 30, alpha=innerRangeAlpha, **fill_betweenParams)
    # Plot median.
    axarr[0, 0].plot(x, x * 25)
    tryToSaveFig('./plotTest')

def getValAtTimeBySimAndSeed(df, compartment, timeToInterpolateAt):
    """Returns a pd.Series of compartment values at timeToInterpolateAt, with multiindex on sim and seed.
    """
    dfValAtTimeBySimeAndSeed = df.groupby(by=['sim','seed']).apply(lambda z: np.interp(timeToInterpolateAt, xp=z['timeSinceSimStarted'], fp=z[compartment]))
    return dfValAtTimeBySimeAndSeed

def getValAtHarvestBySimAndSeed(df, compartment, harvestToGet):
    dfValAtHarvestBySimAndSeed = df.loc[df['nHarvests'] == harvestToGet, ['sim', 'seed', compartment]].groupby(by=['sim','seed'])[compartment].apply(max)
    return dfValAtHarvestBySimAndSeed

@timeMe_decorator
def plotTimeCourse(resultsFileName, figFileName, season, title='', group=0, asPercentages=False, heightAsProportionOfWidth=1, xMax=6, ax=None, rowTitle='', theLegends=None, rowTitleX=-0.5, rowTitleY=0.5, compartments=['S', 'E', 'I', 'D', 'R'],
    plotLabels=[], yLabel=None, RThreshold=np.inf, plotType='individualLines', plotPotentialYield=False, trajectoryToHighlight=None, subplotLetter='', textItems=None, subtitle='', yMax=None, showXLabel=True, showYLabel=True, yTicks=None, annotation=None,
    greyscale=False, yScalingDivisor=None, plotProfitabilityThreshold=None, profitabilityTextAtRight=False, onlyTrajectoryToHighlight=False, subplotLetterXOffset=0.3, subplotLetterYOffset=0.15, markerAtPoint=None, fontSizeTitle=fontSizeTitle, showYTickLabels=True,
    xLabelOffset=0):
    if ax == None:
        fig, ax = plt.subplots()
        saveThisFig = True
    else:
        saveThisFig = False
    if greyscale:
        coloursForThisPlot = defaultdict(lambda: 'k')
    else:
        coloursForThisPlot = colours
    xMin = 0
    assert group in [0,1], "Group must be 0 or 1 (not " + str(group) + ")."
    if group == 1:
        resultsFileName = resultsFileName[:-1] + '1'
    data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileName, season, useYieldResultsFile=False)
    if asPercentages:
        yScalingFactor = 100.0 / nTrees # To change number of trees in a compartment to a percentage of total trees.
        yieldScalingFactor = 100.0 / nTrees
    elif yScalingDivisor != None:
        yScalingFactor = 1.0 / yScalingDivisor
    elif 'detectablePrev' in compartments:
        yScalingFactor = 100.0
    elif set(['treeYield', 'treeYieldcumulative', 'treeYieldLoss', 'treeYieldLosscumulative']) & set(compartments):
        yScalingFactor = 1.0 / nTrees
    else:
        yScalingFactor = 1
        yieldScalingFactor = 1
    theAlpha = 0.05
    # Make the plots
    plotHandles = []
    yMaxSoFar = -9999
    for (indSeries, thisSeries) in enumerate(compartments):
        if len(plotLabels) > indSeries:
            thisPlotLabel = plotLabels[indSeries]
        else:
            thisPlotLabel = ''
        try:
            thisColour = coloursForThisPlot[thisSeries]
        except KeyError as e:
            thisColour = 'k'
        data, colToPlot = hlbutils.combineColumns(data, thisSeries,  RThreshold=RThreshold)
        assert not (data[colToPlot] < 0).any(), "There was at least one negative value to be plotted: " + str(data[['timeSinceSimStarted', colToPlot]].loc[data[colToPlot] < 0])
        dataByTime = pd.pivot_table(data, index=["timeSinceSimStarted"], columns=["sim", "seed"], values=[colToPlot])
        if not onlyTrajectoryToHighlight:
            if plotType == 'functionalBoxplot':
                # Calculate the functional bandwith median.
                dataByTimeAsMatrix = dataByTime.as_matrix()
                functionalBoxplotMedian = stochPlot.getMedian(dataByTimeAsMatrix)
                functionalBoxplotLower, functionalBoxplotUpper = stochPlot.getPercentile(dataByTimeAsMatrix, 0.95)
                # Plot functional boxplot lines.
                ax.fill_between(dataByTime.index, yScalingFactor * functionalBoxplotLower, yScalingFactor * functionalBoxplotUpper, color=coloursForThisPlot[thisSeries], alpha=0.5, **fill_betweenParams)
                ax.plot(dataByTime.index, yScalingFactor * functionalBoxplotMedian, color=coloursForThisPlot[thisSeries])
            elif plotType == 'pointwiseQuantiles':
                # quantile() with multiple q's and axis=1 returns a dataframe with the times as the COLUMN NAMES
                # and the q's as the indices, so have to transpose().
                dataQuantiles = dataByTime.quantile([lowerQuantile, lowerMidQuantile, midQuantile, upperMidQuantile, upperQuantile], axis=1).transpose()
                # Plot range between upper and lower quantiles.
                ax.fill_between(dataQuantiles.index, yScalingFactor * dataQuantiles[lowerQuantile], yScalingFactor * dataQuantiles[upperQuantile], color=coloursForThisPlot[thisSeries], alpha=outerRangeAlpha, **fill_betweenParams)
                ax.fill_between(dataQuantiles.index, yScalingFactor * dataQuantiles[lowerMidQuantile], yScalingFactor * dataQuantiles[upperMidQuantile], color=coloursForThisPlot[thisSeries], alpha=innerRangeAlpha, **fill_betweenParams)
                # Plot median.
                ax.plot(dataQuantiles.index, yScalingFactor * dataQuantiles[midQuantile], color=coloursForThisPlot[thisSeries])
            else: # Plot all lines individually.
                ax.plot(dataByTime.index, yScalingFactor * dataByTime, color=coloursForThisPlot[thisSeries], alpha=theAlpha)
        else:
            if trajectoryToHighlight == None:
                raise ValueError("Have specfied onlyTrajectoryToHighlight=True but not provided a trajectoryToHighlight.")
        if trajectoryToHighlight != None:
            ax.plot(dataByTime.index, yScalingFactor * dataByTime.xs((trajectoryToHighlight['sim'], trajectoryToHighlight['seed']), level=('sim', 'seed'), axis=1), color=coloursForThisPlot[thisSeries])
        # Plot a dot in an area that won't be visible, with the colour for thisCompartmentGroup.
        # Do it here so that we just assign the plot handle for thisCompartmentGroup once,
        # rather than once for every iSim and iSeed.
        thisPlotHandle, = ax.plot(-1,-1, thisColour, label=thisPlotLabel, linewidth=2) # alpha=1 line to use for legend.
        plotHandles.append(thisPlotHandle)
    if plotPotentialYield:
        if ('treeYield' in compartments) or ('treeYieldcumulative' in compartments): # Only makes sense to show potential yield if plotting actual yield.
            # Potential yield is nTrees at time 0, and then increases by nTrees every harvest, which is every time floor(timeSinceSimStarted) increases.
            potentialYield = nTrees * (1.0 + np.floor(dataByTime.index))
            ax.plot(dataByTime.index, yScalingFactor * potentialYield, color=coloursForThisPlot['potentialYield'], zorder=-99)
            thisPlotHandle, = ax.plot(-1, -1, color=coloursForThisPlot['potentialYield'], label=plotLabels[0], linewidth=2)
            plotHandles.insert(0, thisPlotHandle)
    if plotProfitabilityThreshold != None:
        if 'treeYieldLoss' in compartments:
            profitabilityThresholdForPlot = 1 - plotProfitabilityThreshold
        elif 'treeYield' in compartments:
            profitabilityThresholdForPlot = plotProfitabilityThreshold
        ax.plot([0, xMax], profitabilityThresholdForPlot * np.array([1, 1]), color='r', linestyle='--')
        if profitabilityTextAtRight:
            profitabilityTextX = 5.5
        else:
            profitabilityTextX = 0.3
        ax.text(profitabilityTextX, profitabilityThresholdForPlot + 0.05, 'Not profitable', verticalalignment='bottom', fontsize=fontSizePublication, color='r')
        ax.text(profitabilityTextX, profitabilityThresholdForPlot - 0.05, 'Profitable', verticalalignment='top', fontsize=fontSizePublication, color='g')
    if annotation != None:
        ax.text(annotation[1], annotation[2], annotation[0], fontsize=fontSizeTitle)
    if markerAtPoint != None:
        ax.plot(markerAtPoint['x'], markerAtPoint['y'] * yScalingFactor, 'kx')
    if showXLabel:
        hXLabel = ax.set_xlabel("Time (years)", fontsize=fontSizePublication)
    ax.set_xlim([xMin, xMax])
    ax.set_xticks(range(xMax+1))
    if not yTicks is None:
        ax.set_yticks(yTicks)
    # Choose y-label.
    if yLabel == None: # If yLabel != None, whatever the user supplied will be used.
        if 'detectablePrev' in compartments:
            yLabel = 'Prevalence (%)'
        elif compartments == ['detectablePrevInclR']:
            yLabel = 'Prevalence (%) \n(R included)'
        elif compartments == ['treeYieldcumulative']:
            yLabel = 'Cumulative yield'
        else:
            if asPercentages:
                yLabel = 'Trees (%)'
            else:
                yLabel = 'Trees'
    if showYLabel:
        if len(rowTitle) > 0:
            ax.set_ylabel("\n\n\n" + yLabel, fontsize=fontSizePublication)
        else:
            ax.set_ylabel(yLabel, fontsize=fontSizePublication)
        # Set y axis limits.
    if yMax == None:
        if asPercentages or 'detectablePrev' in compartments:
            ax.set_ylim([0, 100])
        elif compartments == ['treeYield']:
            ax.set_ylim([0, 1])
        else:
            ax.set_ylim([0, ax.get_ylim()[1]])
    else:
        ax.set_ylim([0, yMax])
    ax.text(rowTitleX, rowTitleY, rowTitle, transform=ax.transAxes, rotation=90, fontsize=fontSizeTitle)
    #ax.text(0.5, titleY, title, transform=ax.transAxes, fontsize=fontSizeTitle, horizontalalignment='center')
    subplotLetterX = ax.get_xlim()[0] - subplotLetterXOffset * (ax.get_xlim()[1] - ax.get_xlim()[0])
    subplotLetterY = ax.get_ylim()[1] + subplotLetterYOffset * (ax.get_ylim()[1] - ax.get_ylim()[0])
    subplotLetterHandle = ax.text(subplotLetterX, subplotLetterY, subplotLetter, fontsize=fontSizeSubplotLetter)
    ax.text(0.5, 1.3, subtitle, horizontalalignment='center', verticalalignment='bottom', transform=ax.transAxes, fontsize=fontSizePublication)
    print('title: ' + str(title))
    ax.set_title(title + '\n\n', fontsize=fontSizeTitle) # Pad so that the column title, which is placed with ax.text, doesn't go out of the figure.
    ax.tick_params(axis='both', which='major', labelsize=fontSizePublication, length=tickLength)
    if not showYTickLabels:
        ax.set_yticklabels([])
    legendParams = {'handlelength':2, 'ncol':len(plotLabels), 'bbox_to_anchor':(0., 1.02, 1., 1), 'loc':3, 'mode':"expand", 'borderaxespad':0., 'handletextpad':0.25, 'frameon':False, 'fontsize':fontSizePublication}
    legendHandle = ax.legend(plotHandles, plotLabels, **legendParams)
    # Hide borders.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    #ax.spines['bottom'].set_visible(False)
    #ax.spines['left'].set_visible(False)
    # Hide top and right tick marks.
    ax.tick_params(axis='x', which='both', top='off')
    ax.tick_params(axis='y', which='both', right='off')
    if theLegends != None:
        theLegends.append(legendHandle)
    if textItems != None:
        textItems.append(subplotLetterHandle)
    if saveThisFig:
        tryToSaveFig(figFileName, theLegends=[legendHandle], heightAsProportionOfWidth=heightAsProportionOfWidth)
    return data, d, nSims, nSeeds, nTrees

@timeMe_decorator
def plotSeriesAgainstSeries(resultsFileName, figFileName, season, xContents, yContents, profitabilityThreshold, title='', group=0, asPercentages=False, heightAsProportionOfWidth=1, xMax=6, ax=None, rowTitle='', theLegends=None,
    rowTitleX=-0.5, rowTitleY=0.5, yLabel=None, xLabel=None, connectivityParameter=10**(-4), RThreshold=np.inf, plotType='individualLines', subplotLetter='', textItems=None, trajectoryToHighlight=None, actualPrevsToHighlight=None,
    actualPrevsAndPercentilesToSummarise=None, binWidth=0.005, interpolateOverX=False, showXLabel=True, showYLabel=True, showXTickLabels=True, showYTickLabels=True, prev28LabelRotation=90,
    prev28LabelX=29, prev28LabelY=5, xTicks=None, yTicks=None, subplotLetterXOffset=0.30):
    """Generalised plotting function that plots two 'series' (counts of a compartment, counts of sums of compartments, or yield) against each other, aligned on time.

    Args:
        xContents: List containing one or more of 'yield' or compartment names. If there is more than one element, their counts will be added together.
        yContents: List containing one or more of 'yield' or compartment names. If there is more than one element, their counts will be added together.
    """
    # Yield updates once per year, so if either x or y is yield then a few things will work differently. Create a variable to indicate whether yield is being used.
    usingYield = (xContents == 'yield' or yContents == 'yield')
    neededSeries = xContents + yContents # Will use this to extract all necessary series from the results.
    if ax == None:
        fig, ax = plt.subplots()
        saveThisFig = True
    else:
        saveThisFig = False
    xMin = 0
    assert group in [0,1], "Group must be 0 or 1 (not " + str(group) + ")."
    if group == 1:
        resultsFileName = resultsFileName[:-1] + '1'
    if usingYield:
        # Yield updates once per year, so if either x or y is yield then load results from file that only updates once per year.
        data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileName, season, useYieldResultsFile=True)
    else:
        data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileName, season, useYieldResultsFile=False)
    if asPercentages:
        yScalingFactor = 100 / nTrees # To change number of trees in a compartment to a percentage of total trees.
    elif 'treeYield' in [xContents, yContents]:
        yScalingFactor = 1.0 / nTrees
    elif 'treeYieldcumulative' in [xContents, yContents]:
        yScalingFactor = 1.0 / nTrees
    elif 'I+Dcumulative' in [xContents, yContents]:
        yScalingFactor = 1.0 / 10000
    elif 'actualPrev' in yContents:
        yScalingFactor = 100
    else:
        yScalingFactor = 1.0
    if 'detectablePrev' in xContents:
        xScalingFactor = 100
    else:
        xScalingFactor = 1
    theAlpha = 0.05
    xMaxSoFar = -9999
    yMaxSoFar = -9999
    data, xColToPlot = hlbutils.combineColumns(data, xContents, RThreshold=RThreshold, profitabilityThreshold=profitabilityThreshold)
    data, yColToPlot = hlbutils.combineColumns(data, yContents, RThreshold=RThreshold, profitabilityThreshold=profitabilityThreshold)
    # RThreshold doesn't necessarily get applied in combineColumns, so apply it again.
    data = data.loc[data['R'] <= RThreshold]
    data['profitable'] = (1.0 *  data['treeYield'] / nTrees) >= profitabilityThreshold
    dataUnprofitable = data.copy()
    dataUnprofitable.loc[data['profitable'], [xColToPlot, yColToPlot]] = np.nan
    if interpolateOverX:
        xGrid = np.linspace(0, max(data[xColToPlot]), num=1000)
        interpolatedData = hlbutils.interpolateByGroup(data=data, indexVars=['sim', 'seed'], xContents=xColToPlot, yContents=yColToPlot, xGrid=xGrid)
        dataByXContents = pd.pivot_table(interpolatedData, index=[xColToPlot], columns=["sim", "seed"], values=[yColToPlot])
    else:
        dataByTime = pd.pivot_table(data, index=["timeSinceSimStarted"], columns=["sim", "seed"], values=[xColToPlot, yColToPlot])
        dataByXContentsWithNAs = pd.pivot_table(data, index=[xColToPlot], columns=["sim", "seed"], values=[yColToPlot])
        dataByXContents = dataByXContentsWithNAs.interpolate(method='linear', axis=1)
        #dataByXContents = dataByXContentsWithNAs.fillna(method='pad') # Fill values forward over NAs.
        dataUnprofitableByTime = pd.pivot_table(dataUnprofitable, index=["timeSinceSimStarted"], columns=["sim", "seed"], values=[xColToPlot, yColToPlot])
    if plotType == 'binnedQuantiles':
        binStart = 0.00
        binEnd = 0.50 # 50%
        print("For binned quantiles, using bins of width " + str(binWidth))
        binEdges = np.arange(start=binStart, stop=binEnd + 0.01, step=binWidth)
        binCenters = (binEdges[1:] + binEdges[:-1]) / 2.0
        binEdgesForPlot, binnedQuantilesForPlot = hlbutils.getBinnedQuantilesForPlotting(data, binColumn=xContents, quantileColumn=yContents, binEdges=binEdges, quantiles=[0, lowerQuantile, lowerMidQuantile, midQuantile, upperMidQuantile, upperQuantile, 1])
        # Draw light area for outer quantiles.
        ax.fill_between(xScalingFactor * binEdgesForPlot, yScalingFactor * binnedQuantilesForPlot[lowerQuantile], yScalingFactor * binnedQuantilesForPlot[upperQuantile], color='black', alpha=outerRangeAlpha, **fill_betweenParams)
        # Draw dark area for inner quantiles.
        ax.fill_between(xScalingFactor * binEdgesForPlot, yScalingFactor * binnedQuantilesForPlot[lowerMidQuantile], yScalingFactor * binnedQuantilesForPlot[upperMidQuantile], color='black', alpha=innerRangeAlpha, **fill_betweenParams)
        # Plot medians.
        ax.plot(xScalingFactor * binEdgesForPlot, yScalingFactor * binnedQuantilesForPlot[midQuantile], color='black')
        if xContents == 'detectablePrev':
            # Highlight detectable prev. 28% with a line.
            ax.plot([28, 28], [0, 100], color='red', linestyle=':')
            ax.text(prev28LabelX, prev28LabelY, 'Detectable prev.: 28%', rotation=prev28LabelRotation, verticalalignment='bottom', fontsize=8, color='red')
            if yContents == 'actualPrev' and (not actualPrevsToHighlight is None):
                # Do some reporting on where quantiles are.
                for actualPrevThreshold in actualPrevsToHighlight:
                    try:
                        startOfBinWhereQuantileCrossesThreshold, endOfBinWhereQuantileCrossesThreshold, actualPrevWhereQuantileCrossesThreshold = hlbutils.getBinForWhichQuantileWasCrossed(data, binColumn=xContents, yColumn=yContents, binEdges=binEdges, quantileThreshold=upperQuantile, yThreshold=actualPrevThreshold)
                        print('Upper quantile (' + str(upperQuantile * 100) + 'th) crossed actual prevalence threshold ' + str(actualPrevThreshold) + ' in detectable prevalence bin ' + str(startOfBinWhereQuantileCrossesThreshold) + ' - ' + str(endOfBinWhereQuantileCrossesThreshold))
                        # Find out what the actual prev quantile is in the bin where the threshold is crossed. It may be greater than the threshold.
                        ax.plot(xScalingFactor * (startOfBinWhereQuantileCrossesThreshold + endOfBinWhereQuantileCrossesThreshold) / 2.0 * np.array([1, 1]), [0, yScalingFactor * actualPrevWhereQuantileCrossesThreshold], color='red', linestyle='--')
                        if actualPrevThreshold == 0.9:
                            # Include a newline character to stop text overlapping the plotted lines.
                            actualPrevText = 'Actual prev.: \n' + str(int(round(100 * actualPrevThreshold))) + '%'
                        else:
                            actualPrevText = 'Actual prev.: ' + str(int(round(100 * actualPrevThreshold))) + '%'
                        ax.text(startOfBinWhereQuantileCrossesThreshold * xScalingFactor + 1, 5, actualPrevText, rotation=90, verticalalignment='bottom', fontsize=8, color='red')
                    except ValueError:
                        pass
                if not actualPrevsAndPercentilesToSummarise is None:
                    for (actualPrevThreshold, quantile) in actualPrevsAndPercentilesToSummarise:
                        try:
                            startOfBinWhereQuantileCrossesThreshold, endOfBinWhereQuantileCrossesThreshold, _ = hlbutils.getBinForWhichQuantileWasCrossed(data, binColumn=xContents, yColumn=yContents, binEdges=binEdges, quantileThreshold=quantile, yThreshold=actualPrevThreshold)
                            print('Quantile (' + str(quantile  * 100) + 'th) crossed actual prevalence threshold ' + str(actualPrevThreshold) + ' in detectable prevalence bin starting at ' + str(startOfBinWhereQuantileCrossesThreshold) + ' - ' + str(endOfBinWhereQuantileCrossesThreshold))
                        except ValueError:
                            print("Requested quantile " + str(quantile  * 100) + " has not been computed. Please modify code to do so.")
    elif plotType == 'pointwiseQuantiles':
        # quantile() with multiple q's and axis=1 returns a dataframe with the times as the COLUMN NAMES
        # and the q's as the indices, so have to transpose().
        dataQuantiles = dataByXContents.quantile([0, lowerQuantile, lowerMidQuantile, midQuantile, upperMidQuantile, upperQuantile, 1], axis=1).transpose()
        # Plot range between upper and lower quantiles.
        ax.fill_between(xScalingFactor * dataQuantiles.index, yScalingFactor * dataQuantiles[lowerQuantile], yScalingFactor * dataQuantiles[upperQuantile], color='black', alpha=outerRangeAlpha, **fill_betweenParams)
        ax.fill_between(xScalingFactor * dataQuantiles.index, yScalingFactor * dataQuantiles[lowerMidQuantile], yScalingFactor * dataQuantiles[upperMidQuantile], color='black', alpha=innerRangeAlpha, **fill_betweenParams)
        if False: #trajectoryToHighlight != None:
            ax.plot(xScalingFactor * dataByXContents.index, yScalingFactor * dataByXContents.xs((trajectoryToHighlight['sim'], trajectoryToHighlight['seed']), level=('sim', 'seed'), axis=1), color='k')
        else: # Plot median.
            ax.plot(xScalingFactor * dataQuantiles.index, yScalingFactor * dataQuantiles[midQuantile], color='black')
        if xContents == 'detectablePrev':
            ax.plot([28, 28], [0, 100], color='red', linestyle=':')
            ax.text(prev28LabelX, prev28LabelY, 'Detectable prev.: 28%', rotation=prev28LabelRotation, verticalalignment='bottom', fontsize=8, color='red')
            if yContents == 'actualPrev' and actualPrevsToHighlight != None:
                for actualPrevThreshold in actualPrevsToHighlight:
                    try:
                        indexWhereUpperQuantileOfActualPrevCrossesThreshold = [x >= actualPrevThreshold for x in dataQuantiles[upperQuantile]].index(True)
                        ax.plot(dataQuantiles.index[indexWhereUpperQuantileOfActualPrevCrossesThreshold] * xScalingFactor * np.array([1, 1]), [0, 100 * actualPrevThreshold], color='red', linestyle='--')

                        if dataQuantiles.index[indexWhereUpperQuantileOfActualPrevCrossesThreshold] * xScalingFactor < 5:
                            # Include a newline character to stop text overlapping the plotted lines.
                            actualPrevText = 'Actual prev.: \n' + str(int(round(100 * actualPrevThreshold))) + '%'
                        else:
                            actualPrevText = 'Actual prev.: ' + str(int(round(100 * actualPrevThreshold))) + '%'
                        ax.text(dataQuantiles.index[indexWhereUpperQuantileOfActualPrevCrossesThreshold] * xScalingFactor + 1, 5, actualPrevText, rotation=90, verticalalignment='bottom', fontsize=8, color='red')
                        print('Upper quantile (' + str(upperQuantile * 100.0) + 'th) crossed actual prevalence threshold ' + str(actualPrevThreshold) + ' at detectable prevalence ' + str(dataQuantiles.index[indexWhereUpperQuantileOfActualPrevCrossesThreshold] * xScalingFactor))
                    except ValueError:
                        print(str(upperQuantile) + ' quantile of actual prevalence never crossed threshold of ' + str(actualPrevThreshold) + '.')
                        print('Values of ' + str(upperQuantile * 100.0) + ' quantile:')
                        print(dataQuantiles[upperQuantile])
                        print('Max value of ' + str(upperQuantile * 100.0) + ' quantile: ' + str(max(dataQuantiles[upperQuantile])))
                if actualPrevsAndPercentilesToSummarise != None:
                    for (actualPrevThreshold, quantile) in actualPrevsAndPercentilesToSummarise:
                        #raise NotImplementedError('Currently uses getBinForWhichQuantileWasCrossed, but I don\'t think that makes sense when using pointwiseQuantiles.')
                        try:
                            indexUpperQuantileOfActualPrevCrossesThreshold = [x >= actualPrevThreshold for x in dataQuantiles[quantile]].index(True)
                            print('Quantile (' + str(quantile * 100.0) + 'th) crossed actual prevalence threshold ' + str(actualPrevThreshold) + ' when detectable prevalence was ' + str(dataQuantiles.index[indexUpperQuantileOfActualPrevCrossesThreshold] * yScalingFactor) + '%.')
                        except ValueError:
                            print("Requested quantile " + str(quantile * 100.0) + " has not been computed. Please modify code to do so.")
        #print('Actual prev. did not cross threshold ' + str(actualPrevThreshold))
    elif plotType == 'range':
        # Plot range between upper and lower quantiles.
        ax.fill_between(xScalingFactor * dataByXContents.index, yScalingFactor * dataByXContents.max(axis=1), yScalingFactor * dataByXContents.min(axis=1), color='gray')
        # Plot median.
        ax.plot(xScalingFactor * data.loc[(data["sim"]==0) & (data["seed"]==0), xColToPlot], yScalingFactor * data.loc[(data["sim"]==0) & (data["seed"]==0), yColToPlot], color='black')
    else: # Plot all lines individually.
        ax.plot(xScalingFactor * dataByTime[xColToPlot], dataByTime[yColToPlot], alpha=0.05, color='k')
    # Set the labels corresponding to the series types.
    detectablePrevLabel = 'Detectable prevalence (%)'
    actualPrevLabel = 'Actual prevalence (%)'
    detectablePrevInclRLabel = 'Detectable prevalence (incl. R) (%)'
    if asPercentages:
        yieldLabel = 'Yield (%)'
        compartmentLabel = 'Trees (%)'
    else:
        yieldLabel = 'Yield'
        compartmentLabel = 'Trees (N)'
    # Allocate the x and y labels.
    if xLabel == None:
        if xContents == 'treeYield':
            xLabel = yieldLabel
        elif xContents == 'detectablePrev':
            xLabel = detectablePrevLabel
        elif xContents == 'detectablePrevInclR':
            xLabel = detectablePrevInclRLabel
        else:
            xLabel = compartmentLabel
    if yLabel == None:
        if yContents == 'treeYield':
            yLabel = yieldLabel
        elif yContents == 'actualPrev':
            yLabel = actualPrevLabel
        elif yContents == 'treeYieldcumulative':
            yLabel = 'Cumulative yield'
        else:
            yLabel = compartmentLabel
    # Set axes, labels, appearance aspects etc.
    if showXLabel:
        print("showXLabel: " + str(showXLabel))
        ax.set_xlabel(xLabel, fontsize=fontSizePublication)
    if showYLabel:
        print("showYLabel: " + str(showYLabel))
        if len(rowTitle) > 0:
            ax.set_ylabel("\n\n\n\n" + yLabel, fontsize=fontSizePublication) # Padding, to extend the figure out over the row titles.
        else:
            ax.set_ylabel(yLabel, fontsize=fontSizePublication)
    if xMax != None:
        ax.set_xlim([0, xMax])
    elif 'detectablePrev' in xContents:
        ax.set_xlim([0, 30])
    elif asPercentages:
        ax.set_xlim([0, 100])
    elif xContents == 'treeYield':
        ax.set_xlim([0, max(xScalingFactor * 80000, ax.get_ylim()[1])])
    if asPercentages:
        ax.set_ylim([0, 100])
    elif yContents == 'actualPrev':
        ax.set_ylim([0, 100])
    elif yContents == 'treeYield':
        ax.set_ylim([0, max(yScalingFactor * 80000, ax.get_ylim()[1])])
    else:
        ax.set_ylim([0, ax.get_ylim()[1]])# if xContents == 'treeYield':
        xMin, xMax = ax.get_xlim()
        ax.xaxis.set_ticks([xMin, (xMin + xMax) / 2, xMax])
    if not showXTickLabels:
        ax.set_xticklabels([])
    if not showYTickLabels:
        ax.set_yticklabels([])
    if xTicks != None:
        ax.set_xticks(xTicks)
    if not yTicks is None:
        ax.set_yticks(yTicks)
    subplotLetterX = ax.get_xlim()[0] - subplotLetterXOffset * (ax.get_xlim()[1] - ax.get_xlim()[0])
    subplotLetterY = ax.get_ylim()[1] + 0.15 * (ax.get_ylim()[1] - ax.get_ylim()[0])
    subplotLetterHandle = ax.text(subplotLetterX, subplotLetterY, subplotLetter, fontsize=fontSizeSubplotLetter)
    textItems.append(subplotLetterHandle)
    ax.text(rowTitleX, rowTitleY, rowTitle, transform=ax.transAxes, rotation=90, fontsize=fontSizeTitle, horizontalalignment='center')
    ax.tick_params(axis='both', which='major', labelsize=fontSizePublication, length=tickLength)
    ax.set_title(title + '\n\n', fontsize=fontSizeTitle, verticalalignment='bottom') # Pad so that the column title, which is placed with ax.text, doesn't go out of the figure.
    # Hide borders.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    #ax.spines['bottom'].set_visible(False)
    #ax.spines['left'].set_visible(False)
    # Hide top and right tick marks.
    ax.tick_params(axis='x', which='both', top='off')
    ax.tick_params(axis='y', which='both', right='off')
    print('theLegends: ' + str(theLegends))
    if theLegends != None:
        theLegends.append(legendHandle)
    if textItems != None:
        textItems.append(subplotLetterHandle)
    if saveThisFig:
        tryToSaveFig(figFileName, theLegends=[legendHandle], heightAsProportionOfWidth=heightAsProportionOfWidth)
    return data, d, nSims, nSeeds, nTrees

def plotHLBGridOverTime(dataAllSimsSeeds, frameTimes, sim=0, seed=0, ax=None, showScale=True, rotation=0, subplotLetter='', subplotLetterXOffset=0.1, textItems=None, subplotLetterYOffset=0):
    """Plot the trees in their locations, showing their states at each time point.
    """
    if ax == None:
        fig, ax = plt.subplots()
        saveThisFig = True
    else:
        saveThisFig = False
    markerSize = 0.5
    print('Extracting just sim ' + str(sim) + ' and seed ' + str(seed))
    data = dataAllSimsSeeds.loc[(dataAllSimsSeeds['sim'] == sim) & (dataAllSimsSeeds['seed'] == seed)]
    Tinitial = data['Tinitial'].iloc[0]
    lowerLeftY = min(data['y'])
    lowerLeftX = min(data['x'].loc[data['y'] == lowerLeftY])
    lowerRightX = max(data['x'])
    lowerRightY = min(data['y'].loc[data['x'] == lowerRightX])
    if rotation == 'auto':
        # Find location of lower left tree and lower right tree and calculate the angle they form with the horizontal.
        # Can then rotate the grid by this angle to make 'straight'.
        # Consider triangle formed by bottom of grid and horizontal
        rotation = np.arcsin((lowerRightY - lowerLeftY) / distance.euclidean((lowerLeftX,lowerLeftY), (lowerRightX,lowerRightY)))
        print('Calculated rotation: ' + str(rotation) + ' radians.')
    unrotatedX = data['x'].as_matrix()
    unrotatedY = data['y'].as_matrix()
    rotatedX, rotatedY = hlbutils.rotate(unrotatedX, unrotatedY, -rotation, (lowerLeftX, lowerLeftY))
    xToUse = 'rotatedX'
    yToUse = 'rotatedY'
    data[xToUse] = rotatedX
    data[yToUse] = rotatedY
    print('Generating grid for sim ' + str(sim) + ', seed ' + str(seed) + '...')
    for indFrame, thisTime in enumerate(frameTimes):
        print('Making frame ' + str(indFrame+1) + ' of ' + str(len(frameTimes)) + '...')
        # Plot all trees as S, then iterate through compartments, over-plotting trees if they
        # have progressed to that compartment.
        ax.scatter(data[xToUse], data[yToUse], marker='o', color=colours['S'], s=markerSize)
        for thisCompartment in ['E', 'I', 'D', 'R']:
            print('  Plotting compartment ' + thisCompartment + '...')
            ax.scatter(data[xToUse].loc[(data[thisCompartment] - Tinitial) <= thisTime], data[yToUse].loc[(data[thisCompartment] - Tinitial) <= thisTime], marker='o', color=colours[thisCompartment], s=markerSize);
        if thisTime == 1:
            ax.set_title(str(thisTime) + ' year', fontsize=fontSizePublication)
        else:
            ax.set_title(str(thisTime) + ' years', fontsize=fontSizePublication)
    scaleIndicatorLineY = -40
    scaleIndicatorLineX = 350
    scaleIndicatorLineLength = 100
    if showScale:
        ax.plot(scaleIndicatorLineX + np.array([0, scaleIndicatorLineLength]), scaleIndicatorLineY + np.array([0, 0]), color='k', linewidth=2)
        ax.text(scaleIndicatorLineX + scaleIndicatorLineLength/2, scaleIndicatorLineY - 20, str(scaleIndicatorLineLength) + ' m', verticalalignment='top', horizontalalignment='center', fontsize=fontSizePublication)
    ax.set_xlabel('', fontsize=1)
    ax.axis('off')
    ax.tick_params(axis='both', which='major', labelsize=1)
    ax.ticklabel_format(useOffset=False)
    ax.axis('tight')
    ax.set_aspect('equal')
    subplotLetterX = ax.get_xlim()[0] - subplotLetterXOffset * (ax.get_xlim()[1] - ax.get_xlim()[0])
    subplotLetterY = ax.get_ylim()[1] + subplotLetterYOffset
    subplotLetterHandle = ax.text(subplotLetterX, subplotLetterY, subplotLetter, fontsize=fontSizeSubplotLetter)
    if textItems != None:
        textItems.append(subplotLetterHandle)

def plotIndicatorByVarying(resultsFileNames, figFileName, surveysPerYear, xContents, season, profitabilityThreshold, indicatorCompartment, indicatorSummaryFn, RThreshold=np.inf, baseline=None, ax=None, fig=None, rowTitle='', title='',
    rowTitleX=-1.7, rowTitleY=0.5, titleY=1.15, plotType='individualLines', trajectoryToHighlight=None, theLegends=None, textItems=None, yLog=False, subplotLetter='', xLabel='Reduction in infection rates (%)', yMax=None, yMin=0, xMin=0,
    yTicks=None, yLabel=None, yScalingDivisor=None, nCols=1, subplotLetterXOffset=0.1, **kwargs):
    offsetAmount = 2.5
    if len(surveysPerYear) == 3:
        boxplotOffsets = offsetAmount * np.array([-1, 0, 1])
    elif len(surveysPerYear) == 4:
        boxplotOffsets = offsetAmount * np.array([-1, -0.33, 0.33, 1])
    else:
        raise NotImplementedError("boxplotOffsets has not been defined when there are " + str(len(surveysPerYear)) + " different values for surveysPerYear.")
    plotHandles = []
    timeToTakeIndicatorValues = 10
    harvestToTakeIndicatorValues = 11.0
    if indicatorSummaryFn == 'valAtTime':
        indicatorSummaryFnImplementation = lambda x: getValAtTimeBySimAndSeed(x, thisIndicatorCompartment, timeToTakeIndicatorValues)
    elif indicatorSummaryFn == 'valAtHarvest':
        indicatorSummaryFnImplementation = lambda x: getValAtHarvestBySimAndSeed(x, thisIndicatorCompartment, harvestToTakeIndicatorValues)
    elif indicatorSummaryFn == 'max':
        indicatorSummaryFnImplementation = lambda x: max(x[indicatorCompartment])
    if ax == None:
        fig, ax = plt.subplots()
        saveThisFig = True
    else:
        saveThisFig = False
    bgColours = {0: 'green', 1:'red', 2:'blue', 4:'m', 12:'c'} # 'lightred' does not exist; can use 'salmon' instead.
    # Load first data set to check nSims, nSeeds so we can preallocate a matrix.
    data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileNames[(surveysPerYear[0], xContents[0])], season, useYieldResultsFile=False)
    if indicatorCompartment == 'treeYieldcumulative':
        # Total YIELD for a totally healhy orchard over time considered is nTrees * (timeToTakeIndicatorValues + 1), where the +1 is for the harvest at time 0.
        yScalingFactor = 100.0 / ((1 + timeToTakeIndicatorValues) * nTrees)
    elif yScalingDivisor != None:
        yScalingFactor = 1.0 / yScalingDivisor
    elif indicatorCompartment == 'treeYieldLosscumulative':
        yScalingFactor = 1.0 / nTrees # Equivalent harvests.
    elif indicatorCompartment == 'profitYieldcumulative':
        # Total PROFIT for a totally healhy orchard over time considered is ((1 - profitabilityThreshold) * nTrees) * (timeToTakeIndicatorValues + 1), where the +1 is for the harvest at time 0.
        yScalingFactor = 100.0 / ((1 + timeToTakeIndicatorValues) * ((1 - profitabilityThreshold) * nTrees))
    elif indicatorCompartment == 'detectablePrev':
        yScalingFactor = 100.0
    else:
        yScalingFactor = 1.0
    # Load result files and calculate the values for the heatmap.
    for iSurveys, thisSurveys in enumerate(surveysPerYear):
        print('Processing number of surveys ' + str(1 + iSurveys) + '/' + str(len(surveysPerYear)) + '...')
        indicatorVals = -9999 + np.zeros([nSims*nSeeds, len(xContents)])
        for iControl, thisControl in enumerate(xContents):
            data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileNames[(thisSurveys, thisControl)], season, useYieldResultsFile=False)
            #if indicatorCompartment == 'I+Dcumulative':
            #    thisIndicatorCompartment = str(np.round(1 - thisControl/100.0, decimals=2)) + '*' + indicatorCompartment
            #else:
            thisIndicatorCompartment = indicatorCompartment
            print('  Processing primary/secondary control level ' + str(1 + iControl) + '/' + str(len(xContents)) + '. Plotting indicator ' + thisIndicatorCompartment)
            data, colToPlot = hlbutils.combineColumns(data, thisIndicatorCompartment, profitabilityThreshold=profitabilityThreshold, RThreshold=RThreshold)
            indicatorVals[:,iControl] = indicatorSummaryFnImplementation(data)
        assert (indicatorVals >= 0).all(), "There were negative values in indicatorVals."
        # Plot.
        if plotType == 'pointwiseQuantiles':
            # quantile() with multiple q's and axis=1 returns a dataframe with the times as the COLUMN NAMES
            # and the q's as the indices, so have to transpose().
            quantiles = np.percentile(indicatorVals, q=[lowerQuantile, upperQuantile], axis=0)
            # Plot range between upper and lower quantiles.
            ax.fill_between(xContents, yScalingFactor * quantiles[0,:], yScalingFactor * quantiles[1], color=colours[str(thisSurveys)], alpha=0.5, **fill_betweenParams)
        elif plotType == 'boxplot':
            hbox = ax.boxplot(yScalingFactor * indicatorVals, positions=xContents+ boxplotOffsets[iSurveys], whis=[2.5, 97.5], showfliers=False, widths=1, patch_artist=True)
            plt.setp(hbox['medians'], color=colours[str(thisSurveys)], linewidth=2)
            plt.setp(hbox['boxes'], color=bgColours[thisSurveys])
            for box in hbox['boxes']:
                box.set_facecolor('white')
            plt.setp(hbox['whiskers'], color=bgColours[thisSurveys], linestyle='-')
            plt.setp(hbox['caps'], color=bgColours[thisSurveys])
            # For boxplot, can't use set_xticks() normally, so specify which control levels to display and use an empty string for the others.
            #controlBetasEpsilonsToShow = [0, 50, 100]
            #boxplotXTicks = [x if x in controlBetasEpsilonsToShow else '' for x in controlBetasEpsilons]
            boxplotXTicks = range(0, 101, 10)
            ax.set_xticks(boxplotXTicks)
            ax.set_xticklabels(boxplotXTicks)
            ax.set_xlim([xMin - 2 * offsetAmount, 100 + 2 * offsetAmount])
            #bp = ax.boxplot(100.0 * yieldsToPlot / nTrees, positions=xPos, whis='range', whiskerprops=whiskerprops)
        else: # Plot individual trajectories.
            ax.plot(xContents, yScalingFactor * np.transpose(indicatorVals), color=colours[str(thisSurveys)])
            ax.set_xlim([0, ax.get_xlim()[1]])
        if trajectoryToHighlight != None:
            pass
            #ax.plot(dataByTime.index, yScalingFactor * dataByTime.xs((trajectoryToHighlight['sim'], trajectoryToHighlight['seed']), level=('sim', 'seed'), axis=1), color=colours[thisSeries])
        # Plot label as text.
        #ax.text(80, -1 + yScalingFactor * np.interp(80, controlBetasEpsilons, np.max(indicatorVals, axis=0)), str(thisSurveys) + ' surveys per year', horizontalalignment='right', verticalalignment='top', color=colours[str(thisSurveys)], fontsize=fontSizePublication)
        # plot hidden line for legend.
        thisPlotHandle, = ax.plot(-1, -1, color=colours[str(thisSurveys)], label=str(thisSurveys), linewidth=2.5)
        plotHandles.append(thisPlotHandle)
    if indicatorCompartment == 'detectablePrev':
        # Plot 28% threshold.
        ax.plot([0, 100], [28, 28], color='r', linestyle='--')
        ax.text(50, 27, 'Regulatory threshold', verticalalignment='top', horizontalalignment='center', color='r', fontsize=fontSizePublication)
    if yLog:
        ax.set_yscale('log')
    elif yMax != None:
        ax.set_ylim([yMin, yMax])
    if yTicks != None:
        ax.set_yticks(yTicks)
        ax.set_yticklabels(yTicks)
    ax.set_xlabel(xLabel, fontsize=fontSizePublication) # To use Greek letters, can use r'Reduction in $\epsilon$
    if yLabel == None:
        if indicatorCompartment == 'detectablePrev':
            yLabel = 'Max. detectable \nprevalence (%)'
        elif indicatorCompartment == 'treeYieldcumulative':
            yLabel = 'Relative \ncumulative yield (%)'
        elif indicatorCompartment == 'treeYieldLosscumulative':
            yLabel = 'Cumulative yield loss (harvest)'
        elif indicatorCompartment == 'profitYieldcumulative':
            yLabel = 'Relative \ncumulative profit (%)'
        elif indicatorCompartment == 'I+Dcumulative':
            yLabel = 'Adjusted AUC of C+I'
        else:
            yLabel = indicatorCompartment
    ax.set_ylabel(yLabel, fontsize=fontSizePublication)
    subplotLetterX = ax.get_xlim()[0] - subplotLetterXOffset * (ax.get_xlim()[1] - ax.get_xlim()[0])
    if yLog:
        subplotLetterY = 10**(np.log10(ax.get_ylim()[1]) + 0.25 * (np.log10(ax.get_ylim()[1]) - np.log10(ax.get_ylim()[0])))
    else:
        subplotLetterY = ax.get_ylim()[1] + 0.15 * (ax.get_ylim()[1] - ax.get_ylim()[0])
    subplotLetterHandle = ax.text(subplotLetterX, subplotLetterY, subplotLetter, fontsize=fontSizeSubplotLetter)
    textItems.append(subplotLetterHandle)
    # Hide borders.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    #ax.spines['bottom'].set_visible(False)
    #ax.spines['left'].set_visible(False)
    # Hide top and right tick marks.
    ax.tick_params(axis='y', which='minor', left='off')
    ax.tick_params(axis='x', which='both', top='off')
    ax.tick_params(axis='y', which='both', right='off')
    ax.tick_params(axis='both', which='major', labelsize=fontSizePublication, length=tickLength)
    if nCols == 2:
        legendBoxOffset = 0.1
    else:
        legendBoxOffset = 0.3
    legendParams = {'handlelength':1.5, 'ncol':len(plotHandles), 'bbox_to_anchor':(legendBoxOffset, 0.975, 1.0 - 2*legendBoxOffset, 1), 'loc':3, 'mode':"expand", 'borderaxespad':0.,
        'handletextpad':0.25, 'frameon':False, 'fontsize':fontSizePublication, 'columnspacing':0}
    legendHandle = ax.legend(plotHandles, surveysPerYear,  **legendParams)
    # tight_layout does not seem to work for subplots with legend titles, so use the plot title to pad.
    ax.text(0.5, 1.025, 'Surveys per year\n', horizontalalignment='center', verticalalignment='bottom', transform=ax.transAxes, fontsize=fontSizePublication)
    ax.set_title(title + '\n\n', fontsize=fontSizeTitle)
    if theLegends != None:
        theLegends.append(legendHandle)

def plotSimTypeBoxplot(resultsFileNames, figFileName, xContents, season, profitabilityThreshold, indicatorCompartment, indicatorSummaryFn, RThreshold=np.inf, baseline=None, ax=None, fig=None, rowTitle='', title='', rowTitleX=-1.7, rowTitleY=0.5, titleY=1.15,
    plotType='individualLines', trajectoryToHighlight=None, theLegends=None, textItems=None, yLog=False, subplotLetter='', xLabel='Reduction in infection rates (%)', yMax=None, yMin=0, xMin=0, yTicks=None, yLabel=None, yScalingDivisor=None, nCols=1,
    subplotLetterXOffset=0.1, xTicks=None, **kwargs):
    plotHandles = []
    timeToTakeIndicatorValues = 10
    harvestToTakeIndicatorValues = 11.0
    if indicatorSummaryFn == 'valAtTime':
        indicatorSummaryFnImplementation = lambda x: getValAtTimeBySimAndSeed(x, thisIndicatorCompartment, timeToTakeIndicatorValues)
    elif indicatorSummaryFn == 'valAtHarvest':
        indicatorSummaryFnImplementation = lambda x: getValAtHarvestBySimAndSeed(x, thisIndicatorCompartment, harvestToTakeIndicatorValues)
    elif indicatorSummaryFn == 'max':
        indicatorSummaryFnImplementation = lambda x: max(x[indicatorCompartment])
    if ax == None:
        fig, ax = plt.subplots()
        saveThisFig = True
    else:
        saveThisFig = False
    # Load first data set to check nSims, nSeeds so we can preallocate a matrix.
    data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileNames['normal'], season, useYieldResultsFile=False)
    if indicatorCompartment == 'treeYieldcumulative':
        # Total YIELD for a totally healhy orchard over time considered is nTrees * (timeToTakeIndicatorValues + 1), where the +1 is for the harvest at time 0.
        yScalingFactor = 100.0 / ((1 + timeToTakeIndicatorValues) * nTrees)
    elif yScalingDivisor != None:
        yScalingFactor = 1.0 / yScalingDivisor
    elif indicatorCompartment == 'treeYieldLosscumulative':
        yScalingFactor = 1.0 / nTrees # Equivalent harvests.
    elif indicatorCompartment == 'profitYieldcumulative':
        # Total PROFIT for a totally healhy orchard over time considered is ((1 - profitabilityThreshold) * nTrees) * (timeToTakeIndicatorValues + 1), where the +1 is for the harvest at time 0.
        yScalingFactor = 100.0 / ((1 + timeToTakeIndicatorValues) * ((1 - profitabilityThreshold) * nTrees))
    elif indicatorCompartment == 'detectablePrev':
        yScalingFactor = 100.0
    else:
        yScalingFactor = 1.0
    # Load result files and calculate the values for the heatmap.
    indicatorVals = -9999 + np.zeros([nSims*nSeeds, len(xContents)])
    for iXContents, thisXContents in enumerate(xContents):
        data, d, nSims, nSeeds, nTrees, useNextNeighbour = hlbutils.loadResults(resultsFileNames[thisXContents], season, useYieldResultsFile=False)
        #if indicatorCompartment == 'I+Dcumulative':
        #    thisIndicatorCompartment = str(np.round(1 - thisControl/100.0, decimals=2)) + '*' + indicatorCompartment
        #else:
        thisIndicatorCompartment = indicatorCompartment
        data, colToPlot = hlbutils.combineColumns(data, thisIndicatorCompartment, profitabilityThreshold=profitabilityThreshold, RThreshold=RThreshold)
        indicatorVals[:,iXContents] = indicatorSummaryFnImplementation(data)
        #assert (indicatorVals >= 0).all(), "There were negative values in indicatorVals."
    # Plot.
    hbox = ax.boxplot(yScalingFactor * indicatorVals, whis=[2.5, 97.5], showfliers=False, widths=0.5, patch_artist=True)
    plt.setp(hbox['medians'], color='k', linewidth=2)
    plt.setp(hbox['boxes'], color='k')
    for box in hbox['boxes']:
        box.set_facecolor('white')
    plt.setp(hbox['whiskers'], color='k', linestyle='-')
    plt.setp(hbox['caps'], color='k')
    if indicatorCompartment == 'detectablePrev':
        # Plot 28% threshold.
        ax.plot([0, 100], [28, 28], color='r', linestyle='--')
        ax.text(50, 27, 'Regulatory threshold', verticalalignment='top', horizontalalignment='center', color='r', fontsize=fontSizePublication)
    if yLog:
        ax.set_yscale('log')
    elif yMax != None:
        ax.set_ylim([yMin, yMax])
    if xTicks != None:
        ax.set_xticks(range(1, len(xTicks)+1))
        ax.set_xticklabels(xTicks, rotation=45)
    if yTicks != None:
        ax.set_yticks(yTicks)
        ax.set_yticklabels(yTicks)
    ax.set_xlabel(xLabel, fontsize=fontSizePublication) # To use Greek letters, can use r'Reduction in $\epsilon$
    if yLabel == None:
        if indicatorCompartment == 'detectablePrev':
            yLabel = 'Max. detectable \nprevalence (%)'
        elif indicatorCompartment == 'treeYieldcumulative':
            yLabel = 'Relative \ncumulative yield (%)'
        elif indicatorCompartment == 'treeYieldLosscumulative':
            yLabel = 'Cumulative yield loss (harvest)'
        elif indicatorCompartment == 'profitYieldcumulative':
            yLabel = 'Relative \ncumulative profit (%)'
        elif indicatorCompartment == 'I+Dcumulative':
            yLabel = 'Adjusted AUC of C+I'
        else:
            yLabel = indicatorCompartment
    ax.set_ylabel(yLabel, fontsize=fontSizePublication)
    subplotLetterX = ax.get_xlim()[0] - subplotLetterXOffset * (ax.get_xlim()[1] - ax.get_xlim()[0])
    if yLog:
        subplotLetterY = 10**(np.log10(ax.get_ylim()[1]) + 0.25 * (np.log10(ax.get_ylim()[1]) - np.log10(ax.get_ylim()[0])))
    else:
        subplotLetterY = ax.get_ylim()[1] + 0.15 * (ax.get_ylim()[1] - ax.get_ylim()[0])
    subplotLetterHandle = ax.text(subplotLetterX, subplotLetterY, subplotLetter, fontsize=fontSizeSubplotLetter)
    print("Plotted subplot letter " + subplotLetter + " at x: " + str(subplotLetterX) + ", y: " + str(subplotLetterY))
    textItems.append(subplotLetterHandle)
    # Hide borders.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    #ax.spines['bottom'].set_visible(False)
    #ax.spines['left'].set_visible(False)
    # Hide top and right tick marks.
    ax.tick_params(axis='y', which='minor', left='off')
    ax.tick_params(axis='x', which='both', top='off')
    ax.tick_params(axis='y', which='both', right='off')
    ax.tick_params(axis='both', which='major', labelsize=fontSizePublication, length=tickLength)
    # tight_layout does not seem to work for subplots with legend titles, so use the plot title to pad.
    ax.set_title(title + '\n\n', fontsize=fontSizeTitle)
    if theLegends != None:
        theLegends.append(legendHandle)

if __name__ == "__main__":
    """If module is called from command line, just run plotTest.
    """
    plotTest()
