# @authors: Andrew P. Craig
# @email: apc53@cam.ac.uk
# @date: 2017/10/09
# @copyright: MIT License

# Functions for running and processing results of the HLB simulator, as well as some associated utility functions.
# There is some unclear use of 'sim' to mean 'parameter set', and 'seed' to mean 'iteration'.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import glob
from collections import Counter
import scipy
from scipy.integrate import odeint
import subprocess
import math
import platform
from scipy.interpolate import interp1d
import pdb
import bisect
import itertools

def runSims(nParamSets, nSimsPerParamSet, surveysPerYear, MCparamsFile, controlBetaPercent, controlEpsilonPercent, cullRadius, percentageOfSymptomaticsDetected,
    outputDir, RTreesContributeToYield, surveyStartPrevPercent, nGroups, controlBeta1Percent, controlEpsilon1Percent, row, locationFile, plantYear, reuseSeeds=0, crypticPeriodScalingFactor=100,
    exposedPeriodScalingFactor=100, exposedPeriodDistribution=1, simYearsBase=6, makeChangeTimeOutputFile=0, yieldVariationParam=1, primaryInfectionType=0, primaryInfectionDistanceParameter=0,
    crypticPeriodType=1):
    """
        Args:
            exposedPeriodDistribution: 1 (default): Cyclic (non-Weibull);  2: Gamma case;  3: Exponential;  4: Cyclic Weibull
            makeChangeTimeOutputFile: 0: don't make file; 1: make file
    """
    assert nSimsPerParamSet > 0, "nSimsPerParamSet must be 1 or more. Was " + str(nSimsPerParamSet) + "."
    if platform.system() == 'Windows':
        exePrefix = ''
    elif platform.system() == 'Linux':
        exePrefix = './'
    else:
        exePrefix = ''
        Warning('Call to model exe may not be correctly specified for operating system ' + platform.system() + '.')
    assert (nGroups in [1,2]), "No input file exists for groups = " + str(nGroups) + "."
    nTrees = 7700
    if nGroups == 1:
        locationFile = locationFile + '_Groups0'
    elif nGroups == 2:
        locationFile = locationFile + '_Groups01'
    rateEtoCDAdjustment = 1 # Should be 1 unless debugging.
    rateCDToIAdjustment = 1 # Should be 1 unless debugging.
    betaAdjustment = 1 # Should be 1 unless debugging.
    initialInfected = 0 # Should be 0 unless debugging.
    outputIntervalInDays = 10
    seedStart = 1
    rateEtoCD = 1.1 / 365 * rateEtoCDAdjustment # For ages 3-10, mean latent RATE is 1.1.
    rateCDToI = 1.0 / (365.0 * (100.0/500.0)) * rateCDToIAdjustment # Time in cryptic period is gamma distributed, with mean 100/500.
    totalSurveys = 10 * surveysPerYear
    if surveysPerYear == 0:
        revisitIntervals = 14*365
    else:
        revisitIntervals = 365/surveysPerYear
    MCparams = pd.read_csv(MCparamsFile, delimiter = "\s", names=["alpha", "beta", "epsilon", "Tinitial"])
    outputPrefix = outputDir + getOutputFileName(thisRun=row, group='')
    # Run simulator.
    print('    Running simulator...')
    exeArgs = exePrefix + 'HLBParry.exe ' + \
    outputPrefix + ' ' + \
    MCparamsFile + ' ' + str(nParamSets) + ' ' + str(hex(seedStart)) + ' ' + str(exposedPeriodDistribution) + ' ' + str(totalSurveys) + ' ' + \
    'Survey' + str(surveysPerYear) + 'TimesPerYear ' + str(cullRadius) + ' ' + locationFile + ' ' + \
    'Distances ' + str(nTrees) + ' ' + str(initialInfected) + ' ' + str(nSimsPerParamSet) + ' ' + str(controlBetaPercent) + ' ' + str(controlEpsilonPercent) + \
    ' ' + str(crypticPeriodType) + ' ' + str(percentageOfSymptomaticsDetected) + ' ' + outputPrefix + '_Timestep ' + \
    str(outputIntervalInDays) + ' ' + str(RTreesContributeToYield) + ' ' + str(surveyStartPrevPercent) + ' ' + str(controlBeta1Percent) + ' ' + str(controlEpsilon1Percent) + ' ' + \
    str(makeChangeTimeOutputFile) + ' ' + str(simYearsBase) + ' ' + str(reuseSeeds) + ' ' + str(crypticPeriodScalingFactor) + ' ' + str(exposedPeriodScalingFactor) + ' ' + str(plantYear) + ' ' +  \
    str(yieldVariationParam) + ' ' + str(primaryInfectionType) + ' ' + str(primaryInfectionDistanceParameter)
    print('Call: ' + exeArgs)
    try:
        exeOutput = subprocess.check_output(exeArgs, stderr=subprocess.STDOUT, shell=True)
        print(exeOutput)
    except subprocess.CalledProcessError as e:
        print('An error occurred in HLBParry.exe (non-zero exit status). Here is its output:')
        print(e.output)
    return exeArgs


def interpolateByGroup(data, indexVars, xContents, yContents, xGrid):
    """Create a new dataframe that has columns indexVars, xContents, yContents,
    but where xContents are the values of xGrid and yContents are interpolated.
    """
    # Truncate xGrid to range of values that lie within xContents.
    truncatedXGrid = xGrid.copy()
    truncatedXGrid = truncatedXGrid[truncatedXGrid < max(data[xContents])]
    truncatedXGrid = truncatedXGrid[truncatedXGrid > min(data[xContents])]
    print('Max value of xContents: ' + str(max(data[xContents])))
    print('Min value of xContents: ' + str(min(data[xContents])))
    print('Max value of xGrid: ' + str(max(xGrid)))
    print('Min value of xGrid: ' + str(min(xGrid)))
    # Create empty dataframe with columns indexVars, xContents, and empty yContents.
    interpolatedData = pd.DataFrame(columns=(indexVars + [xContents, yContents]))
    interpolatedData[indexVars] = interpolatedData[indexVars].astype('int64')
    uniqueIndexVars = data[indexVars].drop_duplicates()
    # We will iterate over all indexVars, interpolating yContents and adding them into interpolatedData.
    print('Interpolating...')
    for (indRow, thisIndexVars) in uniqueIndexVars.iterrows():
        # Make interpolation function using xContents, yContents for these index variables.
        indicesOfTheseIndexVars = (data[indexVars] == pd.Series(thisIndexVars)).all(axis=1)
        #interpolationFn = interp1d(data[xContents].loc[indicesOfTheseIndexVars], data[yContents].loc[indicesOfTheseIndexVars], fill_value='extrapolate')
        #yContentsAtXGrid = interpolationFn(truncatedXGrid)
        yContentsAtXGrid = np.interp(truncatedXGrid, data[xContents].loc[indicesOfTheseIndexVars], data[yContents].loc[indicesOfTheseIndexVars])
        dictToAppend = {z:([thisIndexVars[z]] * len(truncatedXGrid)) for z in dict(thisIndexVars)}
        dictToAppend.update({ xContents:truncatedXGrid, yContents:yContentsAtXGrid})
        interpolatedData =interpolatedData.append(pd.DataFrame(dictToAppend), ignore_index=True)
    return interpolatedData

def rotate(xs, ys, angle, origin):
    originX, originY = origin # Unpack tuple.
    newxs = originX + math.cos(angle) * (xs - originX) - math.sin(angle) * (ys - originY)
    newys = originY + math.sin(angle) * (xs - originX) + math.cos(angle) * (ys - originY)
    return newxs, newys

def integrateOverTime(grp, compartments):
    return cumtrapzOverTime(grp, compartments)

def cumtrapzOverTime(grp, compartments):
    """Calculates the integral over time of a specified compartment. This exists as a function primarily so that it can be
       used in groupby().apply calls.
    """
    grp[compartments + 'cumulative'] = scipy.integrate.cumtrapz(grp[compartments], grp["timeSinceSimStarted"], axis=0, initial=0)
    return grp

def getInoculumReleasedByYield(grp):
    grp['inoculumReleasedByYield'] = grp['I+Dcumulative'].diff() / grp['treeYieldcumulative'].diff()
    return grp

def D50(df):
    """Calculate the time at which D reaches 50% (of 7700 trees).
    """
    return np.interp(x=(7700/2), xp=df['D'], fp=df['timeSinceSimStarted'])

def combineColumns(df, compartments, RThreshold=np.inf, surveyInterval=(3.0/12), profitabilityThreshold=None):
    # Check whether the compartment already exists, in which case we don't have to do anything.
    if compartments in df.columns:
        return df, compartments
    else:
        if compartments == 'annualIncrementsAUCOfI+D':
            df, _ = combineColumns(df, 'I+Dcumulative', Threshold=RThreshold, surveyInterval=surveyInterval, profitabilityThreshold=profitabilityThreshold)
            df['timeSinceSimStartedRoundedDown'] = df['timeSinceSimStarted'].apply(np.floor)
            df = pd.merge(df, df, left_on=['sim','seed','timeSinceSimStartedRoundedDown'], right_on=['sim','seed','timeSinceSimStartedRoundedDown'], suffixes=['','_AtStartOfYear'], how='left')
            df['annualIncrementsAUCOfI+D'] = df['I+Dcumulative'] - df['I+Dcumulative_AtStartOfYear']
        else:
            if compartments[-len('cumulative'):] == 'cumulative':
                compartmentsNonCumulative = compartments[0:-len('cumulative')]
                cumulative = True
            else:
                compartmentsNonCumulative = compartments
                cumulative = False
            if '*' in compartmentsNonCumulative:
                assert False, "Have decided not to use adjusted AUC."
                scalingFactor = float(compartmentsNonCumulative[0:compartmentsNonCumulative.index('*')])
                compartmentNames = compartmentsNonCumulative[(1 + compartmentsNonCumulative.index('*')):]
            else:
                scalingFactor = 1.0
                compartmentNames = compartmentsNonCumulative
            # Prevalence ignores cumulative-ness.
            if compartments == 'DAccruedDetected':
                # Get D values every three months.
                interpFn = interp1d(df['timeSinceSimStarted'], df['timeSinceSimStarted'])
                timesAtIntervals = interpFn(np.arange(start=0, stop=max(df['timeSinceSimStarted']), step=surveyInterval))
                # Take just the elements every three months, group by (sim, seed), diff, multiply by detectionProb, cumsum.
                DAccruedDetectedAtIntervals = df.loc[df['timeSinceSimStarted'] in timesAtIntervals, ['timeSinceSimStarted', 'D', 'sim', 'seed']] # .pivot(index='timeSinceSimStarted', columns=['sim','seed'], values='D') # .diff().apply(lambda x: multiply(x, detectionProb)).cumsum()
                df.join(DAccruedDetectedAtIntervals, on=['sim','seed','timeSinceSimStarted'], rsuffix='AccruedDetection')
                df.fillna(method='pad')
            elif compartments == 'inoculumReleasedByYield':
                df, _ = combineColumns(df, 'I+D', cumulative=True, RThreshold=RThreshold)
                df, _ = combineColumns(df, 'treeYield', cumulative=True, RThreshold=RThreshold)
                df = df.groupby(['sim', 'seed']).apply(getInoculumReleasedByYield).reset_index()
            elif compartments == 'profitYield':
                # Only counts yield in excess of profitability threshold.
                # For example, if profitability threshold is 0.6 and there are 7700 trees initially,
                # profitYield is max(treeYield - 0.6*7700, 0)
                if profitabilityThreshold == None:
                    raise(ValueError("Trying to calculate profit but profitabilityThreshold has not been set."))
                nTrees = df[['S', 'E', 'I', 'D', 'R']].iloc[0].sum()
                df['profitYield'] = np.maximum(df['treeYield'] - profitabilityThreshold*nTrees, 0)
            elif compartments == 'profitYieldcumulative':
                df, _ = combineColumns(df, 'profitYield', RThreshold=RThreshold, surveyInterval=surveyInterval, profitabilityThreshold=profitabilityThreshold)
                df = df.join(df.groupby(['sim','seed','nHarvests']).last()['profitYield'].groupby(level=['sim','seed']).cumsum(), on=['sim','seed','nHarvests'], rsuffix='cumulative')
            else:
                # Only create column if it doesn't already exist in dataframe.
                if compartments not in df.columns:
                    # Even if cumulative is required, calculate non-cumulative version first.
                    # For example, if compartments is 'I+D', then df['I+D'] will be the sum of df['I'] and df['D'].
                    compartmentsToSum = compartmentNames.split('+')
                    for thisCompartmentToSum in compartmentsToSum:
                        df, _ = combineColumns(df, thisCompartmentToSum, RThreshold=RThreshold, surveyInterval=surveyInterval, profitabilityThreshold=profitabilityThreshold)
                    df[compartmentsNonCumulative] = scalingFactor * df[compartmentsToSum].sum(axis=1)
                    if cumulative:
                        # Integrate over time.
                        df = df.groupby(['sim', 'seed']).apply(lambda x: integrateOverTime(x, compartments=compartmentsNonCumulative)).reset_index()
            # 'Clip' all results after R crosses the specified threshold.
            df.loc[df['R'] > RThreshold] = np.nan
        return df, compartments

def expspace(stop, num):
    """A function to produce an array of exponentially spaced points (the 'opposite' of numpy.logspace).
    Mainly to emphasise upper end of control levels. Always starts at 0.

        Args:
            stop: Last value of returned array.
            num: Number of values in returned array.
    """
    return np.around(stop / np.log(num) * np.log(np.linspace(1, num, num=num+1)))

def cartesianProduct(**kwargs):
    """Example: cartProd(paramA=[1,2], paramB=[3,4]) returns {'paramA':[1,1,2,2], 'paramB':[3,4,3,4]}.
    """
    cartProd = [i for i in itertools.product(*kwargs.values())]
    cartProdUnzipped = zip(*cartProd)
    returnDict = {}
    for iKey, thisKey in enumerate(kwargs.keys()):
        returnDict[thisKey] = cartProdUnzipped[iKey]
    return returnDict

def generateSimSummary(cartesianProductVariables, matchedVariables={}):
    """Wrapper for cartProd that provides default arguments.
    """
    overlap = set(cartesianProductVariables.keys()) & set(matchedVariables.keys())
    if len(overlap) > 0:
        raise ValueError("Variable(s) " + str(overlap) + " appear in both cartesianProductVariables and matchedVariables.keys. This is not allowed.")
    # MCparamsFile, locationFile and plantYear have no defaults, so make sure that they have been specified.
    allSpecifiedVariables = set(cartesianProductVariables.keys()) | set(matchedVariables.keys())
    for requiredVariable in ['MCparamsFile', 'locationFile', 'plantYear']:
        if not requiredVariable in allSpecifiedVariables:
            raise ValueError(requiredVariable + " must be specified, but is not.")
    if ('exposedPeriodScalingFactor' in allSpecifiedVariables) or ('crypticPeriodScalingFactor' in allSpecifiedVariables):
        raise NotImplementedError('Model probably does not handle scaling of exposed period duration or cryptic period duration.')
    # First create the Cartesian product part (all combinations of variables).
    simSummary = cartesianProduct(**cartesianProductVariables)
    # Add on the variables that should take the same values as other variables already in simSummary.
    for key in matchedVariables:
        try:
            simSummary[key] = simSummary[matchedVariables[key]]
        except ValueError as e:
            raise ValueError("Values of matchedVariables must be in cartesianProductVariables, but " + str(matchedVariables[key]) + " is not.")
    # Add on those variables that were not specified, for which defaults are used.
    defaults = {'nParamSets':[1], 'nSimsPerParamSet':[-1], 'surveysPerYear':[0], 'cullRadius':[0], 'controlBeta':[0], 'controlEpsilon':[0], 'RTreesContributeToYield':[0], 'surveyStartPrevPercent':[-1], 'nGroups':[1], 'controlBeta1':[0], 'controlEpsilon1':[0], 'crypticPeriodScalingFactor':[100], 'exposedPeriodScalingFactor':[100], 'exposedPeriodDistribution':[1], 'simDurationYears':[6], 'percentageOfSymptomaticsDetected':[0.5], 'makeChangeTimeOutputFile':[0], 'crypticPeriodType':[1], 'primaryInfectionType':[0], 'primaryInfectionDistanceParameter':[0]}
    # Check how many elements each value in simSummary has, by checking a key that it will definitely contain.
    simSummaryRows = len(simSummary['MCparamsFile'])
    unspecifiedVariables = set(defaults) - allSpecifiedVariables
    for thisUnspecifiedVariable in unspecifiedVariables:
        simSummary[thisUnspecifiedVariable] = defaults[thisUnspecifiedVariable] * simSummaryRows
    return simSummary

def getOutputFileName(thisRun, group):
    return 'output_Run' + str(thisRun) + '_Group' + str(group) # File name previously contained number of surveysPerYear etc., but simplified because it was getting too long.

def queryByDict(df, query, precision=0.001, quiet=True):
    """Returns rows of a pandas.DataFrame with values matching those given in a dict.
    Does approximate matching on float columns.

        Args:
            df: pandas.DataFrame. The data frame to query.
            query: dict.
            precision: numeric. Float variables in query are padded on either side with precision for approximate matching.
            quiet: If False, prints windows on which float variables were matched.
        Returns:
            Rows of df matching query.
    """
    queryDF = pd.DataFrame(query, index=[0])
    # Can't safely match on float columns, so for each float column in query,
    # create a new factor column by binning the values (using pd.cut).
    tmpCols = []
    for col in queryDF.columns:
        if queryDF[col].dtype == float and col in df.columns:
                bins = np.append(-np.inf, [query[col]-precision, query[col]+precision, np.inf])
                out = pd.cut(queryDF[col], bins=bins)
                if not quiet:
                    print('For ' + col + ', matched on window: ' + str(out.cat.categories[1]))
                colAsFactorName = col + '_AS_FACTOR'
                tmpCols.append(colAsFactorName)
                queryDF = queryDF.assign(**{colAsFactorName:out}) # Splat so can use dynamic name.
                # Create a new column in df, using the same cut scheme.
                df = df.assign(**{colAsFactorName:pd.cut(df[col], bins=bins)}) # Splat so can use dynamic name.
                # Drop non-factor version of COL from queryDF.
                queryDF = queryDF.drop(col, axis=1)
    merged = pd.merge(queryDF, df)
    # Drop factor columns.
    merged = merged.drop(tmpCols, axis=1)
    return merged

def getOutputPath(simDir, ID, simSummary, precision=None, **kwargs):
    """Finds the path of the simulation with a specified set of parameters.

        Args:
            simSummary: string. The filename of the simSummary file.
            precision: float. If not None, will be passed on to queryByDict.
            kwargs: dict. Dict of column names and their required values, to be looked for in simSummary.
    """
    if ID == 'test':
        print('Returning path to dummy data created for testing purposes...')
        outputPath = simDir + 'testSims/testData'
    else:
        print('Finding location of required simulation output file...')
        defaults = {'surveysPerYear':0, 'controlBeta':0, 'controlEpsilon':0, 'cullRadius':0, 'nSimsPerParamSet':-1, 'RTreesContributeToYield':0, 'surveyStartPrevPercent':-1, \
        'controlBeta1':0, 'controlEpsilon1':0, 'nGroups':1, 'MCparamsFile':'MCparams_Tinitial1', 'nParamSets':1, \
        'crypticPeriodScalingFactor':100, 'exposedPeriodScalingFactor':100, 'exposedPeriodDistribution':1, 'simDurationYears':6, 'makeChangeTimeOutputFile':0, \
        'primaryInfectionType':0, 'primaryInfectionDistanceParameter':10, 'crypticPeriodType': 1}
        reqdKeysAndVals = defaults.copy()
        reqdKeysAndVals.update(kwargs)
        # Filter simSummary using reqdKeysAndVals using a merge.
        if precision == None: # Use queryByDict's default precision.
            filteredSimSummary = queryByDict(simSummary, reqdKeysAndVals)
        else:
            filteredSimSummary = queryByDict(simSummary, reqdKeysAndVals, precision=precision)
        print('These parameters were specified: ' + str(kwargs))
        print('These parameters were defaults: ' + str({key:defaults[key] for key in defaults if key not in kwargs}))
        keysInKwargsButNotSimSummary = set(kwargs).difference(set(simSummary))
        if len(keysInKwargsButNotSimSummary) > 0:
            print('These parameters were specified but not found in simSummary: ' + str(keysInKwargsButNotSimSummary))
        jobForTheseParams = list(filteredSimSummary['job'])
        runForTheseParams = list(filteredSimSummary['Run'])
        if len(runForTheseParams) != 1:
            raise ValueError('Found ' + str(len(runForTheseParams)) + ' runs meeting criteria. Expected 1.')
        outputDir = 'ID_' + ID + '_Job_' + str(jobForTheseParams[0]) + '/'
        if 'groupFile' in kwargs:
           groupFile = kwargs['groupFile']
        else:
           groupFile = 0 # Default.
        outputPath = simDir + outputDir + getOutputFileName(runForTheseParams[0], groupFile)
        print('Found required output file: ' + str(outputPath))
    return outputPath

def reloadIfModified_decorator(f):
    """Loading and pre-processing the results via loadResults is slow, so this decorator
    checks to see if the file has changed since the last time it was loaded. If not, a cached
    value is returned.
    Separate timestamps and cached values are maintained for each file name.
    """
    fileLastModifiedTime = {} # Keys are outputFileNames.
    cachedReturnVals = {} # Keys are outputFileNames.
    def _inner(*args, **kwargs):
        fileName = args[0]
        fileNewModifiedTime = os.stat(fileName).st_mtime
        if fileName in fileLastModifiedTime and fileName in cachedReturnVals:
            print("File results are cached. Checking if up to date...")
            if fileLastModifiedTime[fileName] >= fileNewModifiedTime:
                print("No output file changes detected.")
            else:
                print("Re-loading results because they have changed since cached.")
                fileLastModifiedTime[fileName] = fileNewModifiedTime
                cachedReturnVals[fileName] = f(*args, **kwargs)
        else:
            print("Loading results because they are not cached.")
            fileLastModifiedTime[fileName] = fileNewModifiedTime
            cachedReturnVals[fileName] = f(*args, **kwargs)
        return cachedReturnVals[fileName]
    return _inner

#@reloadIfModified_decorator
def loadResults(outputFileNames, season, useYieldResultsFile=False):
    useNextNeighbourForYield = True
    if isinstance(outputFileNames, str): # Is a single filename being passed in.
        outputFileNames = [outputFileNames] # If so, make it a len 1 list so the following for loop works.
    for iFileName, thisFileName in enumerate(outputFileNames):
        if useYieldResultsFile:
            timestepOutputFileName = thisFileName + '_Timestep_Yield'
        else:
            # Add 'Timestep' at the end of output filename.
            timestepOutputFileName = thisFileName + '_Timestep'
        thisData = pd.read_csv(timestepOutputFileName, delimiter = "\t")
        thisNSims = len(thisData['sim'].unique())
        thisNSeeds = len(thisData['seed'].unique())
        thisNTrees = thisData[['S','E','I','D','R']].iloc[0].sum()
        thisD = {}
        print("There were " + str(thisNSims) + " parameter sets.")
        if iFileName == 0:
            nSeeds = thisNSeeds
            nSims = thisNSims
            nTrees = thisNTrees
            data = thisData
            d = thisD
        else:
            assert nSeeds == thisNSeeds, "Different number of sims per param set in " + thisFileName + " (" + str(thisNSeeds) + ") from " + outputFileNames[0] + " (" + str(nSeeds) + ")."
            assert nTrees == thisNTrees, "Different number of trees in " + thisFileName + " (" + str(thisNTrees) + ") from " + outputFileNames[0] + " (" + str(nTrees) + ")."
            data = pd.concatenate([data, thisData], axis=0)
            for key in d.keys():
                d[key] = d[key] + thisD[key]
            nSims = nSims + thisNSims
    return data, d, nSims, nSeeds, nTrees, useNextNeighbourForYield

def getBinnedQuantiles(data, binColumn, quantileColumn, binEdges, quantiles):
    """Given (x,y) data, bins x's and returns the quantiles of the corresponding y's for each bin.

    Args:
        data (pandas dataframe):
        binColumn (string): Name of x column, which will be binned.
        quantileColumn (string): Name of y column, of which quantiles will be taken.
        binEdges (numpy array): The edges of the bins.
        quantiles (list): The quantiles that will be taken.

    Returns:
        quantilesCorrespondingToBinEdgesForPlot:
    """
    nValuesInEachBin = data.groupby(pd.cut(data[binColumn], binEdges)).size()
    dataQuantiles = data.groupby(pd.cut(data[binColumn], binEdges)).quantile(quantiles)[quantileColumn]
    quantilesCorrespondingToBins = {}
    for thisQuantile in quantiles:
        quantilesCorrespondingToBins[thisQuantile] = dataQuantiles.iloc[dataQuantiles.index.get_level_values(1) == thisQuantile]
    return quantilesCorrespondingToBins

def getBinnedQuantilesForPlotting(data, binColumn, quantileColumn, binEdges, quantiles):
    quantilesCorrespondingToBins = getBinnedQuantiles(data=data, binColumn=binColumn, quantileColumn=quantileColumn, binEdges=binEdges, quantiles=quantiles)
    binEdgesForPlot = np.hstack((binEdges[0], np.repeat(binEdges[1:-1], 2), binEdges[-1]))
    quantilesCorrespondingToBinEdgesForPlot = {}
    for thisQuantile in quantiles:
        quantilesCorrespondingToBinEdgesForPlot[thisQuantile] = np.repeat(quantilesCorrespondingToBins[thisQuantile], 2)
    return binEdgesForPlot, quantilesCorrespondingToBinEdgesForPlot

def getBinForWhichQuantileWasCrossed(data, binColumn, yColumn, binEdges, quantileThreshold, yThreshold):
    """Given (x,y) data, bins x's, gets quantiles of the corresponding y's, and returns the bin at which a particular quantile threshold on a y threshold was met.
    """
    quantilesCorrespondingToBins = getBinnedQuantiles(data=data, binColumn=binColumn, quantileColumn=yColumn, binEdges=binEdges, quantiles=[quantileThreshold])
    isQuantileAboveThreshold = [x >= yThreshold for x in quantilesCorrespondingToBins[quantileThreshold]]
    indexOfBinWhereQuantileCrossesThreshold = isQuantileAboveThreshold.index(True)
    startOfBinWhereQuantileCrossesThreshold = binEdges[indexOfBinWhereQuantileCrossesThreshold]
    endOfBinWhereQuantileCrossesThreshold = binEdges[1 + indexOfBinWhereQuantileCrossesThreshold]
    yWhereQuantileCrossesThreshold = quantilesCorrespondingToBins[quantileThreshold][indexOfBinWhereQuantileCrossesThreshold]
    return startOfBinWhereQuantileCrossesThreshold, endOfBinWhereQuantileCrossesThreshold, yWhereQuantileCrossesThreshold
