//---------------------------------------------------------------------------//
// Copyright (c) 2017 Matthew Parry & Andrew P. Craig
//
// Distributed under The MIT License (MIT)
// See accompanying file LICENSE.txt
//---------------------------------------------------------------------------//

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cmath>
#include <algorithm>
#include <sys/stat.h>
#include <boost/math/special_functions/gamma.hpp>
#include "random_numbers.hpp"
#include "utils.hpp"
#include "yield.hpp"
#ifndef _WIN32
#define strcpy_s(a,b,c) strcpy(a,c)
#endif

using namespace std;
using namespace boost::math;

// Simulation code uses Sellke algorithm to get exposure times; E->I and I->D transitions occur independently
// for each tree using chosen latent period model and hardwired cryptic period model respectively;
// both these latter transitions are age-dependent.

// Below is specific to block in question
// N gives number of trees, Dtot number of symptomatics observed over n_survey (see below) sweeps
// Z is size of vector of all possible distances; recall trees have been snapped to 1m x 1m grid
// Zx is width in x direction of grid
// distances found by lookup not computation.
int N;
const int Dtot=305;
const int Z=240779;
const int Zx=583;

// T1 is start of control; T2 is end of control, when psyllids are assumed eradicated
// infection rates go linearly to zero in interval [T1,T2].
const double Tfinal=99;
const double Tmax=99;
const double T1=99;
const double T2=101;

// Transition parameters for 2-3 month I->D.
const double T_ID_i=0.17;
const double T_ID_f=0.25;
const double lambda_ID=100.0;
const double nu_ID=500.0;

// Yield parameters.
const double yield_calculation_time_interval = 1;

// Year zero.
const int zero_year = 0;

struct pressure {
  double zero;
  double dot;
};

struct model {
  double alpha;
  double beta;
  double epsilon;
  double Tinitial;
  int subregionID;
};

// Prototype some functions.

double binomial_distribution_sample(int n,double p);

// Computes the kernel function.
double SGkernel (model m, double r);

// Finds which tree has the next event and when.
bool SGmin(double *dt, int * p_i, double * p_dt, double * dt_cull);

// After an event, "pulls down" all times-to-next-event by dt and all exposure thresholds.
void SGpull(double * p_dt, double t, double t0, double Ti, int *s, double *y, double *z, double *dt, double *dt_cull);

// "Pushes out" changes in infectious pressure when a tree becomes infectious or is removed.
void SGpush(int * p_i, double t, double Ti, timeline *T, int *s, double *y, double *z, double *dt, model m, double *ker, double *beta_by_b);

// Determines what removal time a symptomatic tree will have given removal schedule.
double SGremove(double t, double *detects, double *removes, int n_survey, double detection_prob);

// Determines what sweep a symptomatic tree will be detected in given sweep schedule.
int SGcycle(double t, double *detects);

// This is the function of time that scales the infectious rates.
double SGs (double t);

// This is the integral of the previous function; needed to compute the accumulated infectious challenge.
double SGS (double t, double Ti);

// Inverse of the previous function.
double SGSinv (double z, double Ti);

void writeOutput(double &next_output_time, double t, double T_sim_until, int nGroups, double * detectable_prev, int * current_S, int * current_E, int * current_I, int * current_D, int * current_R, double detection_prob, double * actual_prev, double &next_yield_calculation_time, int &n_harvests, double * total_yield, double * E, double * R, double * severity_s0, double * severity_rL, bool R_trees_contribute_to_yield, timeline * T, double * yield_variation_param, double * cumulative_total_yield, double * total_yield_loss, double * total_yield_if_uninfected, double * cumulative_total_yield_loss, std::ofstream * yield_out, int j, int i_iteration_for_this_param_set, model &m, bool surveying_has_started, double &next_output_time_less_Tinitial, int this_seed, std::ofstream * SGtimestep_out, double output_interval);

int main (int argc,char* argv[])
{

	// Define some variables
	timeline *T;
	int *s; // s is type of next event for each tree.
	double *E,*I,*D,*R,*y,*z,*dt,*dt_cull,*severity_s0,*severity_rL,*beta_by_tree,*control_beta,*control_epsilon,*yield_variation_param;
	double *r,*ker;
	double *total_yield; // Total in the sense that it is the total of all trees at that instant.
	double *cumulative_total_yield;
	double *total_yield_if_uninfected; // What yield would be if all trees were S.
	double *total_yield_loss;
	double *cumulative_total_yield_loss;
	double *detectable_prev;
	double *actual_prev;
	int *current_S,*current_E,*current_I,*current_D,*current_R;
	double *primary_infection_by_tree,*distance_function_by_tree;
	int n_trees_in_group[] = {0,0};
	int n_survey;

	char line[80],ctmp[80],outfile[200],infile[80],survey_file[80],timeline_file[80];
	string timestep_outfile_prefix, timestep_outfile_full;
	unsigned long initial_seed;
	int M;
	int * p_i;
	int i_min,i,j,k,it,c,i_age;
	double dt_min,t,t0,dmin,dmax,age;
	double * p_dt;
	double rmv;
	model m;
	int EI_model;
	double cull_radius;
	bool cull_flag;
	double remove_time_base;
	int prev_s;
	int prev_i_min; // Used for debugging.
	int n_initial_infected; // Used for debugging.
	int cryptic_period_type; // 1 if gamma distributed, 2 if exponentially distributed
	double detection_prob; // Probability that a tree in state D is detected if a survey occurs. Was implicitly 1 in Parry paper.
	int iterations_per_param_set;
	bool reuse_seeds; // If true, seeds will be re-used for each parameter set. If false, each parameter set will use different seeds.
	int temp_iterations_per_param_set;
	double next_output_time,output_interval,next_output_time_less_Tinitial; // Track next_output_time_less_Tinitial separately rather than calculating, to hopefully improve precision.
	double next_yield_calculation_time;
	int n_harvests;
	bool R_trees_contribute_to_yield; // This is used to check relative contributions to yield.
	// Theoretically, setting R_trees_contribute_to_yield to true would correspond to instantly replanting rogued trees in a far-away grove, where they continue to contribute yield but do not contribute to new infections.
	int survey_start_prev_percent; // Set to negative to use survey input file. Set to positive to start surveying at time that prevalence is reached.
	bool surveying_has_started;
	double found_D, found_D_prev_percent;
	int nGroups;
	bool make_change_time_output_file;
	int T_sim_years; // Number of years that will be simulated (started from Tinitial).
	double T_sim_until; // Time that last event must exceed (incorportating Tinitial and output_interval).
	int this_seed;
	double nu_scaling_factor; // Can be used to change second parameter (nu) of gamma distribution used for cryptic period.
	double e_scaling_factor; // Can be used to change exposed period.
	int plant_year_all_trees; // Can be used to control the age of all trees.
	double yield_variation_metaparam;
	int primary_infection_type; // 0 for 'normal'; 1 for primary infection at edge trees only.
	int primary_infection_distance_parameter; 

	N = stoi(argv[11]); // Number of trees

	// Allocate some memory

	T = new timeline[N];
	s = new int[N];
	E = new double[N];
	I = new double[N];
	D = new double[N];
	R = new double[N];
	y = new double[N];
	z = new double[N];
	dt = new double[N];
	r = new double[N * N];
	ker = new double[N * N]; // Will hold distances between each tree.
	dt_cull = new double[N];
	severity_s0 = new double[N];
	severity_rL = new double[N];
	beta_by_tree = new double[N];
	yield_variation_param = new double[N];
	primary_infection_by_tree = new double[N];
	distance_function_by_tree = new double[N];
	control_beta = new double[2]; // There should be nGroups control_beta variables, as nGroups may be 1, so possibly creating a redundant one. 
	control_epsilon = new double[2]; // There should be nGroups control_beta variables, as nGroups may be 1, so possibly creating a redundant one. 
	total_yield = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 
	cumulative_total_yield = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 
	total_yield_if_uninfected = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 
	total_yield_loss = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 
	cumulative_total_yield_loss = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 
	detectable_prev = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 
	actual_prev = new double[2]; // There should be nGroups total_yield variables, as nGroups may be 1, so possibly creating a redundant one. 

	// Load runtime parameters from SGparams.

	// argv[0] is .exe name.
	strcpy_s(outfile,200,argv[1]);
	ofstream SGout;
	make_change_time_output_file = stoi(argv[24]) == 1;
	if (make_change_time_output_file) {
		SGout.open(outfile);
	}
	strcpy_s(infile,79,argv[2]);
	ifstream SGin;
	SGin.open(infile);
	if (!SGin) {
		cout << "Could not find MCparams file " << infile << ". Exiting.\n";
		return EXIT_FAILURE;
	}
	M = stoi(argv[3]);
	strcpy_s(ctmp,79,argv[4]);
	initial_seed = strtol(ctmp,NULL,0);
	EI_model = stoi(argv[5]);
	n_survey = stoi(argv[6]);
	strcpy_s(survey_file,79,argv[7]);
	cull_radius = strtod(argv[8],NULL);
	strcpy_s(timeline_file,79,argv[9]);
	// argv[10] is distance_file; no longer used but there still needs to be a placeholder argument.
	// argv[11] is number of trees.
	n_initial_infected = stoi(argv[12]);
	temp_iterations_per_param_set = stoi(argv[13]);
	if(temp_iterations_per_param_set >= 1) {
		iterations_per_param_set = temp_iterations_per_param_set;
		reuse_seeds = true;
	}
	else {
		iterations_per_param_set = 1;
		reuse_seeds = false;
	}
	control_beta[0] = stod(argv[14], NULL) / 100.0;
	control_epsilon[0]  = stod(argv[15], NULL) / 100.0;
	cryptic_period_type = stoi(argv[16]);
	detection_prob = stoi(argv[17]) / 100.0;
	// argv[18] was output_timeline_prefix.
	output_interval = stoi(argv[19]) / 365.0;
	R_trees_contribute_to_yield = stoi(argv[20]) == 1; // False in real-world situtations. Set to true to check contributions to yield.
	survey_start_prev_percent = stoi(argv[21]);  // Set to negative to use survey input file. Set to positive to start surveying at time that prevalence is reached.
	control_beta[1] = stod(argv[22], NULL) / 100.0;
	control_epsilon[1] = stod(argv[23], NULL) / 100.0;
	// argv[24] is whether to produce a change time output file.
	T_sim_years = stoi(argv[25]); //
	reuse_seeds = stoi(argv[26]) == 1;
	nu_scaling_factor = stoi(argv[27]) / 100.0;
	e_scaling_factor = stoi(argv[28]) / 100.0;
	plant_year_all_trees = stoi(argv[29]);
	yield_variation_metaparam =  strtod(argv[30], NULL);
	primary_infection_type = stoi(argv[31]);
	primary_infection_distance_parameter = stoi(argv[32]);

	// Display parameters extracted from arguments.
	cout << "Output file (change times):              " << outfile << "\n";
	cout << "Input file:                              " << infile << "\n";
	cout << "Number of parameter sets:                " << M << "\n";
	cout << "Initial seed:                            " << initial_seed << "\n";
	cout << "Latent period type:                      " << EI_model << "\n";
	cout << "Number of surveys:                       " << n_survey << "\n";
	cout << "Survey file:                             " << survey_file << "\n";
	cout << "Cull radius:                             " << cull_radius << "\n";
	cout << "Timeline file:                           " << timeline_file << "\n";
	cout << "Number of initial infected:              " << n_initial_infected << "\n";
	cout << "Number of trees:                         " << N << "\n";
	cout << "Number of iterations per param set:      " << iterations_per_param_set << "\n";
	cout << "Beta control (group 0):                  " << control_beta[0] << "\n";
	cout << "Epsilon control (group 0):               " << control_epsilon[0]<< "\n";
	cout << "Cryptic period type:                     " << cryptic_period_type << "\n";
	cout << "Detection probability:                   " << detection_prob << "\n";
	//cout << "Output file prefix (time steps):       " << timestep_outfile_prefix << "\n";
	cout << "Output time step (years):                " << output_interval << "\n";
	cout << "R trees contribute to yield:             " << R_trees_contribute_to_yield << "\n";
	cout << "Survey start prevalence:                 " << survey_start_prev_percent << "\n";
	cout << "Beta control (group 1):                  " << control_beta[1] << "\n";
	cout << "Epsilon control (group 1):               " << control_epsilon[1] << "\n";
	cout << "Reuse seeds:                             " << reuse_seeds << "\n";
	cout << "Nu scaling factor (for gamma dist):      " << nu_scaling_factor << "\n";
	cout << "E scaling factor:                        " << e_scaling_factor << "\n";
	cout << "Year in which trees were planted:        " << plant_year_all_trees << "\n";
	cout << "Yield variation parameter:               " << yield_variation_metaparam << "\n";
	cout << "Primary infection type:                  " << primary_infection_type << "\n";
	cout << "Primary infection distance parameter:    " << primary_infection_distance_parameter << "\n";

	ifstream SGtimeline;
	// This file gives location and timeline for each tree.
	SGtimeline.open(timeline_file);
	if (!SGtimeline) {
		cout << "Could not find timeline file " << timeline_file << ". Exiting.\n";
		return EXIT_FAILURE;
	} else {
		SGtimeline.getline(line, 80);
		if (strcmp(line, "block x y planted group dist_to_edge") != 0) {
			cout << "Header of timeline file " << timeline_file << " must be: 'block x y planted group dist_to_edge'";
			return EXIT_FAILURE;
		}
		nGroups = 0;
		for(i = 0; i < N; i++) {
			SGtimeline >> T[i].block >> T[i].x >> T[i].y >> T[i].age >> T[i].group >> T[i].distance_from_edge;
			T[i].age = (double)plant_year_all_trees + 0.25 - zero_year;
			nGroups = max(nGroups,1 + T[i].group);
			++n_trees_in_group[T[i].group];
		}
		SGtimeline.close();
	}

	current_S = new int[nGroups];
	current_E = new int[nGroups];
	current_I = new int[nGroups];
	current_D = new int[nGroups];
	current_R = new int[nGroups];

	// Create one timestep_out output file per group.
	ofstream *SGtimestep_out;
	SGtimestep_out = new ofstream[nGroups];
	ofstream *yield_out;
	yield_out = new ofstream[nGroups];
	for(int iGroup = 0; iGroup < nGroups; iGroup++) {
		std::string timestep_output_file_name = outfile + std::to_string(iGroup) + "_Timestep";
		SGtimestep_out[iGroup].open(timestep_output_file_name);
		// Write header.
		SGtimestep_out[iGroup] << "sim\tseed\ttime\tS\tE\tI\tD\tR\ttreeYield\tTinitial\tsurveyingHasStarted\ttimeSinceSimStarted\tactualSeed\ttreeYieldcumulative\tnHarvests\ttreeYieldLoss\ttreeYieldLosscumulative\tdetectablePrev\tactualPrev\n";
		std::string yield_output_file_name = outfile + std::to_string(iGroup) + "_Timestep_Yield";
		yield_out[iGroup].open(yield_output_file_name);
		// Write header.
		yield_out[iGroup] << "sim\tseed\ttime\tS\tE\tI\tD\tR\ttreeYield\tTinitial\tsurveyingHasStarted\ttimeSinceSimStarted\tactualSeed\ttreeYieldcumulative\tnHarvests\ttreeYieldLoss\ttreeYieldLosscumulative\tdetectablePrev\tactualPrev\n";
	}

	// Precompute distances between each pair of trees.
	for (int i_tree = 0; i_tree < N; i_tree++) {
		for (int i_other_tree = 0; i_other_tree < N; i_other_tree++) {
			// Euclidean distance.
			r[i_tree + N * i_other_tree] = sqrt(pow(T[i_tree].x - T[i_other_tree].x, 2) + pow(T[i_tree].y - T[i_other_tree].y, 2));
		}
	}

	// Initialize surveys.
	double *detects_base = new double[n_survey + 1]; 
	double *removes_base = new double[n_survey + 1];
	double *detects = new double[n_survey + 1];
	double *removes = new double[n_survey + 1];
	detects_base[n_survey] = Tfinal + Tmax; removes_base[n_survey] = Tfinal + Tmax + 1; // For events after Tfinal.
	if(survey_start_prev_percent <= 0) { // Survey at times specified in survey file.
		ifstream SGsurvey;
		SGsurvey.open(survey_file);
		if (!SGsurvey) {
			cout << "Could not find survey file " << survey_file << ". Exiting.\n";
			return EXIT_FAILURE;
		}
		else {
			SGsurvey.getline(line,80);
			for(i = 0; i < n_survey; i++) {
				SGsurvey >> detects_base[i] >> removes_base[i];
			}
			SGsurvey.close();
		}
	}

	// Initialize rate of infection and threshold.

	this_seed = initial_seed - 1; // -1 because it will be incremented before first use, if different seed is used for every param set/iteration combination.
	for(j = 0; j < M; j++) {
		cout << "Running parameter set (from sample of posterior) " << j + 1 << " of " << M << "\n";

		// Read in parameters from Markov chain.

		SGin >> m.alpha >> m.beta >> m.epsilon >> m.Tinitial >> m.subregionID;
		cout << "alpha is " << m.alpha << ", beta is " << m.beta << ", epsilon is " << m.epsilon << ", Tinitial is " << m.Tinitial << " (from subregion ID " << m.subregionID << ")\n";

		for(int i_tree = 0; i_tree < N; i_tree++) {
			for (int i_other_tree = 0; i_other_tree < N; i_other_tree++) {
				ker[i_tree + N * i_other_tree] = SGkernel(m, r[i_tree + N * i_other_tree]);
			}
		}
		for(i = 0; i < N; i++) {
			beta_by_tree[i] = m.beta * (1 - control_beta[T[i].group]);
		}

		// Calculate primary infectious pressure for each tree.
		if (primary_infection_type == 0) {
			// Primary infection is 'normal' (same for all trees).
			for (int i_tree = 0; i_tree < N; i_tree++) {
				primary_infection_by_tree[i_tree] = m.epsilon;
			}
		} else {
			// Idea is to allocate primary infection such that the same amount of primary infection is allocated as it would be if primary infection were.
			// 'normal' (not based on distance from edge), but restricted to only those trees close to the edge.
			// 1. Sum the 'edgeness' (function of distance from edge) over all trees.
			double total_distance_function = 0;
			for (int i_tree = 0; i_tree < N; i_tree++) {
				// Step function: 1 if less than primary_infection_distance_parameter, 0 otherwise.
				distance_function_by_tree[i_tree] = (T[i_tree].distance_from_edge <= primary_infection_distance_parameter);
				total_distance_function += distance_function_by_tree[i_tree];
			}
			// 2. Calculate how much primary infection would have been allocated, if primary infection were 'normal' (not based on distance from edge).
			double total_primary_infection = N * m.epsilon;
			// 3. Normalise such that this amount of primary infection is allocated, but based on distance from edge.
			for (int i_tree = 0; i_tree < N; i_tree++) {
				primary_infection_by_tree[i_tree] = distance_function_by_tree[i_tree] * total_primary_infection / total_distance_function;
			}
		}

		// Reseed random number generator and set up removal times.

		for(int i_iteration_for_this_param_set = 0; i_iteration_for_this_param_set < iterations_per_param_set; i_iteration_for_this_param_set++) {
			if(reuse_seeds) {
				cout << "  Running simulation " << i_iteration_for_this_param_set + 1 << " of " << iterations_per_param_set << " for this parameter set. nth simulation of each parameter set has same seed.\n";
				this_seed = initial_seed + i_iteration_for_this_param_set;
			}
			else {
				cout << "  Running simulation " << i_iteration_for_this_param_set + 1 << " of " << iterations_per_param_set << " for this parameter set (using previously unused seed).\n";
				++this_seed;
			}
			init_genrand(this_seed);

			t = m.Tinitial;
			cout << "    Initial time (Tinitial) is " << m.Tinitial << ".\n";

			// Need at least T_sim_years of simulation, starting from m.Tinitial. But also need sim to extend another output_interval
			// years after that, to make sure that there is an output on or after (m.Tinitial + T_sim_years).
			T_sim_until = m.Tinitial + T_sim_years + output_interval; 

			p_i = &i_min;
			p_dt = &dt_min;

			// Initialise each tree.
			for(i = 0; i < N; i++) {
				
				// Redundant initial parametrization of exposure times etc for each tree.
				E[i] = Tmax + 1;
				I[i] = Tmax + 2;
				D[i] = Tmax + 3;
				R[i] = Tmax + 4;

				// Find yield variation parameter for each tree.
				// Generate random multiplier: Uniform random in range [1 - yield_variation_metaparam, 1 + yield_variation_metaparam].
				yield_variation_param[i] = ru() * (yield_variation_metaparam * 2) + 1 - yield_variation_metaparam; 
				

				// s is type of most recent event for each tree.
				s[i] = 0;
				// y is infectious pressure for each tree.
				y[i] = primary_infection_by_tree[i] * (1 - control_epsilon[T[i].group]);
				// z is exposure threshold for each tree.
				z[i] = -log(ru());
				// dt is time to next event in current state of orchard.
				if(n_initial_infected == 0) {
					dt[i] = SGSinv(z[i] / y[i],m.Tinitial) - m.Tinitial;
				}
				else {
					if(m.epsilon > 0) {
						throw "Cannot have episilon (primary infection rate) > 0 and n_initial_infected > 0.";
					}
					else {
						if(n_initial_infected != 1) {
							throw "n_initial_infected must be 0 or 1.";
						}
						else {
							i_min = 0; // Initially infected tree is tree 0.
							E[i_min] = t; // Set the exposure time, for the tree that has the next event, to the next event time.
							s[i_min] = 1; // Set the event type (S->E), for the tree that has the next event.
							age = t - T[i_min].age; // Determine age.
							i_age = (age > 3) + (age > 10); // Determine age category, used to find latent period parameter.
							dt[i_min] = e_scaling_factor * SGEI(EI_model,E[i_min],A_EI[i_age],A2_EI[i_age],lambda_EI[i_age],nu_EI[i_age],A3_EI[i_age]); // Scaled by e_scaling_factor.
							SGpush(p_i,t,m.Tinitial,T,s,y,z,dt,m,ker,beta_by_tree);
							severity_s0[i_min] = get_severity_s0(age);
							severity_rL[i_min] = get_severity_rL(age);
						}
					}
				}

				// Set culling times to be after Tfinal (will be brought forward if a tree is culled-by-radius).
				dt_cull[i] = Tfinal + 1;

			}

			i_min = 0; // The tree that will have the next event.
			dt_min = 0; // The time until the next event.

			// First event; will be exposure since all trees start susceptible.
			t0 = t;
			SGmin(dt,p_i,p_dt,dt_cull); // Finds which tree has the next event and when.
			t = t + dt_min; // First event time.
			k = s[i_min]; // First event type.

			// Initialise group data.
			for(int iGroup = 0; iGroup < nGroups; iGroup++) {
				if(iGroup == 0) {
					current_S[0] = n_trees_in_group[0] - n_initial_infected;
				}
				else {
					current_S[iGroup] = n_trees_in_group[iGroup];
				}
				current_E[iGroup] = n_initial_infected;
				current_I[iGroup] = 0;
				current_D[iGroup] = 0;
				current_R[iGroup] = 0;
			}
			next_output_time = m.Tinitial;
			next_output_time_less_Tinitial = 0;
			next_yield_calculation_time = m.Tinitial;
			n_harvests = 0; // Will be incremented just before first output.

			for(int iGroup = 0; iGroup < nGroups; iGroup++) {
				total_yield[iGroup] = calculate_yield(N,E,R,severity_s0,severity_rL,next_output_time,R_trees_contribute_to_yield,iGroup,T,yield_variation_param);
				cumulative_total_yield[iGroup] = 0; // Initialise to 0. Will be updated with calculated yield just before pre-simulation output.
				total_yield_if_uninfected[iGroup] = calculate_yield_if_uninfected(N,iGroup,T,yield_variation_param);
				total_yield_loss[iGroup] = total_yield_if_uninfected[iGroup] - total_yield[iGroup];
				cumulative_total_yield_loss[iGroup] = 0; // Initialise to 0. Will be updated just before pre-simulation output.
			}

			// Initialise survey data.
			if(survey_start_prev_percent > 0) { // Start surveying when prevalence reaches specified level. Time unknown at this point, so initialise all surveys to after Tfinal.
				for(i = 0; i <= n_survey; i++) {
					detects[i] = Tfinal + Tmax;  removes[i] = Tfinal + Tmax + 1; 
				}
				surveying_has_started = false;
			}
			else {
				surveying_has_started = true;
				// Set detection and removal times to be offset by Tinitial, so that they always occur at the same times after the start of the simulation.
				for(i = 0; i <= n_survey; i++) {
					detects[i] = detects_base[i] + m.Tinitial;
					removes[i] = removes_base[i] + m.Tinitial;
				}
			}

			// Do initial output.
			writeOutput(next_output_time, t, T_sim_until, nGroups, detectable_prev, current_S, current_E, current_I, current_D, current_R, detection_prob, actual_prev, next_yield_calculation_time, n_harvests, total_yield, E, R, severity_s0, severity_rL, R_trees_contribute_to_yield, T, yield_variation_param, cumulative_total_yield, total_yield_loss, total_yield_if_uninfected, cumulative_total_yield_loss, yield_out, j, i_iteration_for_this_param_set, m, surveying_has_started, next_output_time_less_Tinitial, this_seed, SGtimestep_out, output_interval);

			// Main simulation loop.
			while(t < T_sim_until)
			{

				// cout << "t: " << t << "\n";
				switch(k) {

					// S->E transition
				case 0:
					E[i_min] = t; // Set the exposure time, for the tree that has the next event, to the next event time.
					SGpull(p_dt,t,t0,m.Tinitial,s,y,z,dt,dt_cull); // after an event, "pulls down" all times-to-next-event by dt and all exposure thresholds.
					s[i_min] = 1; // Set the event type (S->E), for the tree that has the next event.
					age = t - T[i_min].age; // Determine age.
					i_age = (age > 3) + (age > 10); // Determine age category, used to find latent period parameter.
					dt[i_min] = e_scaling_factor * SGEI(EI_model,E[i_min],A_EI[i_age],A2_EI[i_age],lambda_EI[i_age],nu_EI[i_age],A3_EI[i_age]); // Scaled by e_scaling_factor.
					--current_S[T[i_min].group];
					++current_E[T[i_min].group];
					severity_s0[i_min] = get_severity_s0(age);
					severity_rL[i_min] = get_severity_rL(age);
					break;

					// E->I transition.
				case 1:
					I[i_min] = t; // Set the infectious time, for the tree that has the next event, to the next event time.
					SGpull(p_dt,t,t0,m.Tinitial,s,y,z,dt,dt_cull); // After an event, "pulls down" all times-to-next-event by dt and all exposure thresholds.
					s[i_min] = 2; // Set the event type (E->I), for the tree that has the next event.
					if(cryptic_period_type == 1) {
						dt[i_min] = gamma_p_inv(lambda_ID,ru()) / (nu_ID * nu_scaling_factor); // Find the time until the next transition (gamma distributed).
					}
					else if(cryptic_period_type == 2) {
						dt[i_min] = gamma_p_inv(1,ru()) / ((nu_ID * nu_scaling_factor) / lambda_ID); // Find the time until the next transition (exponential distributed).
					} 
					else if (cryptic_period_type == 3) {
						age = t - T[i_min].age; // Determine age.
						i_age = (age > 3) + (age > 10); // Determine age category, used to find cryptic period parameter.
						dt[i_min] = cryptic_period_constant[i_age]; // Find the time until the next transition (constant).
					}
					else {
						throw "cryptic_period_type must be 1 (gamma distribution), 2 (exponential distribution) or 3 (constant).";
					}
					SGpush(p_i,t,m.Tinitial,T,s,y,z,dt,m,ker,beta_by_tree); // "Pushes out" changes in infectious pressure when a tree becomes infectious or is removed.
					--current_E[T[i_min].group];
					++current_I[T[i_min].group];
					break;

					// I->D transition.
				case 2:
					D[i_min] = t; // Set the detection(able?) time, for the tree that has the next event, to the next event time.
					SGpull(p_dt,t,t0,m.Tinitial,s,y,z,dt,dt_cull); // After an event, "pulls down" all times-to-next-event by dt and all exposure thresholds.
					s[i_min] = 3; // Set the event type (I->D), for the tree that has the next event.
					remove_time_base = SGremove(t,detects,removes,n_survey,detection_prob) - t;
					dt[i_min] = remove_time_base + (1e-6)*ru(); // Determines what removal time a symptomatic tree will have given removal schedule.
					// Cull trees within cull_radius.
					if(cull_radius > 0) {
						for(int jTree = 0; jTree < N; jTree++) {
							if(jTree != i) { // Don't check itself.
								if(s[jTree] < 3) { // If j is S, E or I.
									if(cull_radius >= sqrt(pow(T[i_min].x - T[jTree].x,2) + pow(T[i_min].y - T[jTree].y,2))) {  // If within cull_radius (Euclidean distance).
										dt_cull[jTree] = remove_time_base + (1e-6)*ru(); // Determines what removal time a tree will have given removal schedule.
									}
								}
							}
						}
					}
					--current_I[T[i_min].group];
					++current_D[T[i_min].group];
					break;

					// D->R transition.
				case 3:
					R[i_min] = t; // Set the removed time, for the tree that has the next event, to the next event time.
					SGpull(p_dt,t,t0,m.Tinitial,s,y,z,dt,dt_cull); // After an event, "pulls down" all times-to-next-event by dt and all exposure thresholds.
					prev_s = s[i_min];
					s[i_min] = 4; // Set the event type (D->R), for the tree that has the next event.
					dt[i_min] = Tfinal + 1; // This tree has no more SEIR events.
					dt_cull[i_min] = Tfinal + 1; // This tree has no more cull events.
					if(prev_s >= 2) { // "Push out" change in infectious pressure only if tree was previously infectious.
						SGpush(p_i,t,m.Tinitial,T,s,y,z,dt,m,ker,beta_by_tree); // "Pushes out" changes in infectious pressure when a tree becomes infectious or is removed.
					}
					switch(prev_s) {
					case 0:
						--current_S[T[i_min].group];
						break;
					case 1:
						--current_E[T[i_min].group];
						break;
					case 2:
						--current_I[T[i_min].group];
						break;
					case 3:
						--current_D[T[i_min].group];
						break;
					}
					++current_R[T[i_min].group];
				}

				// If prevalence has reached level required to start surveying, start surveying.
				// D_prev_percent = 100.0 * current_D / (N - current_R);
				// Each D tree will be found with probability detection_prob, so number of diseased trees found is given by a binomial distribution with n=current_D, p=detection_prob.
				if(survey_start_prev_percent > 0) {
					found_D = binomial_distribution_sample(current_D[0],detection_prob);
					// Find 'prevalence' using found_D.
					found_D_prev_percent = 100.0 * found_D / (1.0 * N);
					if(survey_start_prev_percent <= found_D_prev_percent && !surveying_has_started) {
						surveying_has_started = true;
						// First survey is time t + years between surveys.
						// In this case we don't need to adjust by Tinitial, because that is effectively included in t.
						for(i = 0; i < n_survey; i++) {
							detects[i] = t + (i + 1) * (1.0 / (n_survey / 10.0)); // n_survey is surveys in 10 years, so n_survey/10.0 is surveys per year, and 1.0/(n_survey/10.0) is years between surveys.
							removes[i] = detects[i];
						}
						// Allocate trees that are already D to a future survey.
						for(i = 0; i < N; i++) {
							if(D[i] <= t) { // If the tree is D... (Includes tree that just became D.)
								remove_time_base = SGremove(t,detects,removes,n_survey,detection_prob) - t;
								dt[i] = remove_time_base + (1e-6)*ru(); // determines what removal time a symptomatic tree will have given removal schedule
							}
						}
					}
				}

				t0 = t;
				prev_i_min = i_min;
				cull_flag = SGmin(dt,p_i,p_dt,dt_cull);  // Finds which tree has the next event and when. Modifies p_i and p_dt (and therefore i_min and dt_min also).
				t = t + dt_min;
				if(cull_flag) {
					k = 3;
				}
				else {
					k = s[i_min];
				}

				// Output current state if required.
				// Outputting while next_output_time < t means that there should be an output each time t crosses an output time,
				// which means that there will be no more events before that output time. 
				writeOutput(next_output_time, t, T_sim_until, nGroups, detectable_prev, current_S, current_E, current_I, current_D, current_R, detection_prob, actual_prev, next_yield_calculation_time, n_harvests, total_yield, E, R, severity_s0, severity_rL, R_trees_contribute_to_yield, T, yield_variation_param, cumulative_total_yield, total_yield_loss, total_yield_if_uninfected, cumulative_total_yield_loss, yield_out, j, i_iteration_for_this_param_set, m, surveying_has_started, next_output_time_less_Tinitial, this_seed, SGtimestep_out, output_interval);
			} 

			// Now decide when trees actually detected and removed, if that occurs
			// and save results.
			if(make_change_time_output_file) {
				for(i = 0; i < N; i++) {
					if(s[i] >= 3 && n_survey > 0 && D[i] <= detects[n_survey - 1]) // If tree entered R or D and detection time is before (or simultaneous to) last detection survey.
					{
						it = SGcycle(D[i],detects); // Determines what sweep a symptomatic tree will be detected in given sweep schedule.
						if(it == 0) // Set dmin to previous sweep time (or -3.0 if there is no previous sweep).
						{
							dmin = -3.0;
						}
						else
						{
							dmin = detects[it - 1];
						}
						dmax = detects[it];
						rmv = removes[it];
						c = it + 1;
					}
					else // Tree was S, E or I, or entered D after last survey.
					{
						dmin = Tfinal + 1.0 / 365.0;
						dmax = 10.0;
						rmv = dmax + 1;
						c = 0;
					}
					SGout << T[i].block << "\t" << c << "\t" << T[i].x << "\t" << T[i].y << "\t" << dmin << "\t" << dmax << "\t" << rmv << "\t" << T[i].age + zero_year - 0.25 << "\t" << j << "\t" << i << "\t" << E[i] << "\t" << I[i] << "\t" << D[i] << "\t" << R[i] << "\t" << m.Tinitial << "\t" << i_iteration_for_this_param_set << "\t" << T[i].group << "\n";
				}
			}

		}
	}
	// Close input and output streams.

	SGin.close();

	if(make_change_time_output_file) {
		SGout.close();
	}

	for(int iGroup = 0; iGroup < nGroups; iGroup++) {
		SGtimestep_out[iGroup].close();
		yield_out[iGroup].close();
	}

	delete[] T;
	delete[] s;
	delete[] E;
	delete[] I;
	delete[] D;
	delete[] R;
	delete[] y;
	delete[] z;
	delete[] dt;
	delete[] r;
	delete[] ker;
	delete[] dt_cull;
	delete[] detects;
	delete[] removes;
	delete[] detects_base;
	delete[] removes_base;
	delete[] severity_s0;
	delete[] severity_rL;
	delete[] beta_by_tree;
	delete[] control_beta;
	delete[] control_epsilon;
	delete[] current_S;
	delete[] current_E;
	delete[] current_I;
	delete[] current_D;
	delete[] current_R;
	delete[] total_yield;
	delete[] cumulative_total_yield;
	delete[] primary_infection_by_tree;
	delete[] distance_function_by_tree;

	return 0;
}

void writeOutput(double &next_output_time, double t, double T_sim_until, int nGroups, double * detectable_prev, int * current_S, int * current_E, int * current_I, int * current_D, int * current_R, double detection_prob, double * actual_prev, double &next_yield_calculation_time, int &n_harvests, double * total_yield, double * E, double * R, double * severity_s0, double * severity_rL, bool R_trees_contribute_to_yield, timeline * T, double * yield_variation_param, double * cumulative_total_yield, double * total_yield_loss, double * total_yield_if_uninfected, double * cumulative_total_yield_loss, std::ofstream * yield_out, int j, int i_iteration_for_this_param_set, model &m, bool surveying_has_started, double &next_output_time_less_Tinitial, int this_seed, std::ofstream * SGtimestep_out, double output_interval)
{
	while (next_output_time < t && next_output_time < T_sim_until) {
		// Update actual and detectable prevalence.
		for (int iGroup = 0; iGroup < nGroups; iGroup++) {
			detectable_prev[iGroup] = calculate_detectable_prev(current_S[iGroup], current_E[iGroup], current_I[iGroup], current_D[iGroup], current_R[iGroup], detection_prob);
			actual_prev[iGroup] = calculate_actual_prev(current_S[iGroup], current_E[iGroup], current_I[iGroup], current_D[iGroup], current_R[iGroup]);
		}
		// Get yield if required.
		if (next_output_time >= next_yield_calculation_time) {
			n_harvests++;
			next_yield_calculation_time += yield_calculation_time_interval;
			for (int iGroup = 0; iGroup < nGroups; iGroup++) {
				total_yield[iGroup] = calculate_yield(N, E, R, severity_s0, severity_rL, next_output_time, R_trees_contribute_to_yield, iGroup, T, yield_variation_param);
				cumulative_total_yield[iGroup] += total_yield[iGroup];
				total_yield_loss[iGroup] = total_yield_if_uninfected[iGroup] - total_yield[iGroup];
				cumulative_total_yield_loss[iGroup] += total_yield_loss[iGroup];
				yield_out[iGroup] << j << "\t" << i_iteration_for_this_param_set << "\t" << next_output_time << "\t" << current_S[iGroup] << "\t" << current_E[iGroup] << "\t" << current_I[iGroup] << "\t" << current_D[iGroup] << "\t" << current_R[iGroup] << "\t" << total_yield[iGroup] << "\t" << m.Tinitial << "\t" << surveying_has_started << "\t" << next_output_time_less_Tinitial << "\t" << this_seed << "\t" << cumulative_total_yield[iGroup] << "\t" << n_harvests << "\t" << total_yield_loss[iGroup] << "\t" << cumulative_total_yield_loss[iGroup] << "\t" << detectable_prev[iGroup] << "\t" << actual_prev[iGroup] << "\n";
			}
		}
		for (int iGroup = 0; iGroup < nGroups; iGroup++) {
			SGtimestep_out[iGroup] << j << "\t" << i_iteration_for_this_param_set << "\t" << next_output_time << "\t" << current_S[iGroup] << "\t" << current_E[iGroup] << "\t" << current_I[iGroup] << "\t" << current_D[iGroup] << "\t" << current_R[iGroup] << "\t" << total_yield[iGroup] << "\t" << m.Tinitial << "\t" << surveying_has_started << "\t" << next_output_time_less_Tinitial << "\t" << this_seed << "\t" << cumulative_total_yield[iGroup] << "\t" << n_harvests << "\t" << total_yield_loss[iGroup] << "\t" << cumulative_total_yield_loss[iGroup] << "\t" << detectable_prev[iGroup] << "\t" << actual_prev[iGroup] << "\n";
		}
		next_output_time += output_interval;
		next_output_time_less_Tinitial += output_interval;
	}
}

double binomial_distribution_sample(int n, double p) {
	int successes = 0;
	for(int i = 0; i < n; i++) {
		if(ru() < p) {
			++successes;
		}
	}
	return successes;
}

double SGkernel (model m, double r)
{
  return (exp(-r/m.alpha));  
}

// Finds which tree has the next event and when.
bool SGmin (double *dt,int * p_i,double * p_dt,double * dt_cull)
{
	bool cull_flag = false;
	*p_i = 0;
	*p_dt = dt[0];
	for(int i = 1; i < N; i++) { // Find smallest dt.
		if(dt[i] < *p_dt || dt_cull[i] < *p_dt)
		{
			*p_i = i;
			if(dt[i] < dt_cull[i]) { // Next event is SEID event.
				*p_dt = dt[i];
				cull_flag = false;
			}
			else {
				*p_dt = dt_cull[i]; // Next event is cull-by-radius event.
				cull_flag = true;
			}
		}
	}
	return cull_flag;
}

// After an event, "pulls down" all times-to-next-event by dt and all exposure thresholds.
void SGpull(double * p_dt,double t,double t0,double Ti,int *s,double *y,double *z,double *dt,double *dt_cull)
{
	double S1,S0,s1,s0;
	S1 = SGS(t,Ti); // Integral of \rho(t) in the paper; Ti is time of psyllid arrival.
	S0 = SGS(t0,Ti); // Integral of \rho(t) in the paper; Ti is time of psyllid arrival.
	s1 = SGs(t); // rho(t) in Parry et al. 2014.
	s0 = SGs(t0); // rho(t) in Parry et al. 2014.
	for(int i = 0; i < N; i++) {
		dt[i] = dt[i] - *p_dt; // "Pulls down" all times-to-next-event by dt (SEID).
		dt_cull[i] = dt_cull[i] - *p_dt; // "Pulls down" all times-to-next-event by dt (cull-by-radius).
		if(s[i] == 0 && t < T2)
		{
			z[i] = z[i] - (S1 - S0) / s0*y[i];
			y[i] = s1 / s0*y[i];
		}
	}
}

// "Pushes out" changes in infectious pressure when a tree becomes infectious or is removed.
void SGpush(int * p_i,double t,double Ti,timeline *T,int *s,double *y,double *z,double *dt,model m,double *ker,double *beta_by_tree)
{
	double tmp,S1,s1;
	S1 = SGS(t,Ti); // Integral of \rho(t) in the paper; Ti is time of psyllid arrival
	s1 = SGs(t); // rho(t) in Parry et al. 2014.
	for(int i = 0; i < N; i++) {
		tmp = ker[i + N * (*p_i)];
		if(s[i] == 0 && t < T2)
		{
			if(s[*p_i] == 2) // Next event is E -> I
			{
				y[i] = y[i] + s1*beta_by_tree[*p_i] / (m.alpha*m.alpha)*tmp;
			}
			else if(s[*p_i] == 4) // Next event is D -> R
			{
				y[i] = max(0.0, y[i] - s1*beta_by_tree[*p_i] / (m.alpha*m.alpha)*tmp); // Max with 0 to avoid negative y, which can otherwise occur due to numerical instability.
			}
			dt[i] = SGSinv(z[i] * s1 / y[i] + S1,Ti) - t;
		}
	}
}

double SGremove(double t,double *detects,double *removes,int n_survey,double detection_prob)
{
	int itime;
	double random_number;
	itime = 0;
	random_number = ru();
	while(
		itime < n_survey && // Detects has n_survey+1 elements, so last element of detects is detects[n_survey].
		(t>detects[itime] // Keep going if t > detects[itime] (i.e., if that detection round has already passed), or
		|| random_number > detection_prob)) // if detection_prob is exceeded (i.e., if inspector misses symptomatic tree).
	{
		itime = itime + 1;
		random_number = ru();
	}
	return (removes[itime]);
}

int SGcycle(double t, double *detects)
{
  int itime;
  itime=0;
  while (t>detects[itime]) // Finds the first element of detects that is greater than or equal to t.
    {
      itime=itime+1;
    }
  return (itime); // Returns the index of the first element of detects that is greater than or equal to t.
}

inline
double SGs (double t)
{
  if (t<T1)
    return (1);
  else if (t>=T2)
    return (0);
  else
  return ((T2-t)/(T2-T1));
}

inline
double SGS (double t, double Ti) // The integral of SGs (\rho from the paper), with Ti subtracted.
{
  if (t<T1)
    return (t-Ti);
  else if (t>=T2)
    return (0.5*(T1+T2)-Ti);
  else
  return (t-Ti-0.5*(t-T1)*(t-T1)/(T2-T1));
}

double SGSinv (double z, double Ti)
{
  if (z<T1-Ti)
    return (z+Ti);
  else if (z>=0.5*(T1+T2)-Ti)
    return (Tfinal+1);
  else
  return (T2-(T2-T1)*sqrt(1-2*(z-(T1-Ti))/(T2-T1)));
}

