//---------------------------------------------------------------------------//
// Copyright (c) 2017 Matthew Parry & Andrew P. Craig
//
// Distributed under The MIT License (MIT)
// See accompanying file LICENSE.txt
//---------------------------------------------------------------------------//

#pragma once
#ifndef UTILS_H
#define UTILS_H
#include "random_numbers.hpp"
#include <boost/math/special_functions/gamma.hpp>

using namespace boost::math;

const double Pi=3.14159265358979;

// Transitions parameters for various latent period models.
const double lambda_EI[]={12.0,14.0,10.0};
const double nu_EI[]={15.0,8.0,4.0};
const double A_EI[]={2.4,1.1,0.8};
const double A2_EI[]={4.6,0.43,0.17};
const double A3_EI[]={0.041,0.041,0.041}; // 15 days; 0.041 = 15 / 365 years
// cryptic_period_constant is non-value for trees less than 3 years old;
// for others, it is: Mean time spent in exposed compartment, plus mean time spent in cyptic compartment, minus 15 days.
// Mean times are in paper supplementary material.
const double cryptic_period_constant[] = { -999,0.92 + 0.2 - 0.041,1.26 + 0.2 - 0.041 }; 

struct timeline {
  int block;
  int x;
  int y;
  double Dmin;
  double Dmax;
  double R;
  double age;
  int group;
  double distance_from_edge;
};

int test_me() {
	return 1;
}

double calculate_detectable_prev(int current_S,int current_E,int current_I,int current_D,int current_R,double detection_prob) {
	double numerator = detection_prob * current_D;
	int denominator = current_S + current_E + current_I + current_D;
	// Handle denonimator separately to avoid dividing by zero.
	if(0 == denominator) {
		// Denominator must be greater than or equal to numerator, so if it is zero then so is numerator.
		return 0;
	}
	else {
		return ((double)numerator) / ((double)denominator);
	}
}

double calculate_actual_prev(int current_S,int current_E,int current_I,int current_D,int current_R) {
	int numerator =  current_E + current_I + current_D;
	int denominator = current_S + current_E + current_I + current_D;
	// Handle denonimator separately to avoid dividing by zero.
	if(0 == denominator) {
		// Denominator must be greater than or equal to numerator, so if it is zero then so is numerator.
		return 0;
	}
	else {
		return ((double)numerator) / ((double)denominator);
	}
}

inline
double H (double T, double E, double k, double A, double t0)
{
  return ((k+A)*T-A/(2*Pi)*(sin(2*Pi*(T+E-t0))-sin(2*Pi*(E-t0))));
}

// For cyclic latent period; uses H.
double SGFinv (double r,double E,double k,double A,double t0)
{
	double err = 1.0;
	double tol = 1e-6;
	int side = 0;
	int ctr = 0;
	int max_iter = 500;
	double d,x,x0,x1,y,y0,y1;

	if(r == 0)
		x = 0;
	else
	{
		d = -log(1 - r);
		x0 = 0;
		y0 = -d;
		x1 = d / k;
		y1 = H(x1,E,k,A,t0) - d;
		err = 1.0;
		while(err > tol && ctr < max_iter)
		{
			ctr = ctr + 1;
			x = (y1*x0 - y0*x1) / (y1 - y0);
			y = H(x,E,k,A,t0) - d;
			if(y*y0 > 0)
			{
				x0 = x;
				y0 = y;
				if(side == -1)
					y1 = y1 / 2;
				side = -1;
			}
			else
			{
				x1 = x;
				y1 = y;
				if(side == 1)
					y0 = y0 / 2;
				side = 1;
			}
			err = abs(x1 - x0);
		}
		if(ctr >= max_iter)
		{
			// Do nothing.
		}
		x = 0.5*(x0 + x1);
	}
	return (x);
}

inline
double H2 (double T, double E, double k, double A, double t0)
{
  return (k*T+2*A*(T*T*T/6-T*T/(4*Pi)*sin(2*Pi*(T+E-t0))-T/(4*Pi*Pi)*cos(2*Pi*(T+E-t0))+1.0/(8*Pi*Pi*Pi)*(sin(2*Pi*(T+E-t0))-sin(2*Pi*E-t0))));
}

// For cyclic Weibull latent period (called cycgam in params file for no good reason); uses H2.
double SGF2inv (double r, double E, double k, double A, double t0)
{
  double err=1.0;
  double tol=1e-6;
  int side=0;
  int ctr=0;
  int max_iter=500;
  double d, x, x0, x1, y, y0, y1;

  if (r==0)
    x=0;
  else
    {
      d=-log(1-r);
      x0=0;
      y0=-d;
      x1=d/k;
      y1=H2(x1,E,k,A,t0)-d;
      err=1.0;
      while (err>tol && ctr < max_iter)
	{
	  ctr=ctr+1;
	  x=(y1*x0-y0*x1)/(y1-y0);
	  y=H2(x,E,k,A,t0)-d;
	  if (y*y0>0)
	    {
	      x0=x;
	      y0=y;
	      if (side==-1)
		  y1=y1/2;
	      side=-1;
	    }
	  else
	    {
	      x1=x;
	      y1=y;
	      if (side==1)
		  y0=y0/2;
	      side=1;
	    }
	  err=abs(x1-x0);
	}     
      if (ctr>=max_iter)
	{
	  // Do nothing.
	}
      x=0.5*(x0+x1);
    }
  return (x);
}

// Samples an infectious time for an exposed tree; uses SGFInv, SGF2inv.
double SGEI (int EI_model,double E,double A,double A2,double lambda,double nu,double A3)
{
	double dt;
	switch(EI_model) {

	case 5: // Fixed value.
		dt = A3;
		break;

	case 4: // Cyclic Weibull.
		dt = SGF2inv(ru(),E,1e-6,A2,0.0);
		break;

	case 3: // Exponential.
		dt = -log(ru()) / A;
		break;

	case 2: // Gamma.
		dt = gamma_p_inv(lambda,ru()) / nu;
		break;

	default: // Cyclic (non-Weibull).
		dt = SGFinv(ru(),E,1e-6,A,0.0);

	}
	return dt;
}
#endif